#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.utils import iconParts
from appy.model.fields.group import Group
from appy.model.workflow.state import State
from appy.model.workflow import emptyDict, Role

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
UNTRIGGERABLE = 'Transition "%s" on %s can\'t be triggered.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Transition:
    '''Represents a workflow transition'''
    class Error(Exception): pass
    traverse = {}

    def __init__(self, states, condition=True, action=None, show=True,
                 confirm=False, group=None, icon=None, sicon=None,
                 redirect=None, historizeActionMessage=False):
        # In its simpler form, "states" is a list of 2 states:
        # (fromState, toState). But it can also be a list of several
        # (fromState, toState) sub-lists. This way, you may define only 1
        # transition at several places in the state-transition diagram. It may
        # be useful for "undo" transitions, for example.
        self.states = self.standardiseStates(states)
        self.condition = condition
        if isinstance(condition, str):
            # The condition specifies the name of a role
            self.condition = Role(condition)
        self.action = action
        self.show = show # If False, the end user will not be able to trigger
        # the transition. It will only be possible by code.
        self.confirm = confirm # If True, a confirm popup will show up.
        self.group = Group.get(group)
        # If you want to specify an icon being in folder <yourApp>/static,
        # specify "<yourApp>/<yourImage>". Else, default icon "appy/action.svg"
        # will be used.
        self.icon, self.iconBase = iconParts(icon or 'action.svg')
        # You may specify, in attribute "sicon", an alternate icon suitable when
        # rendered as a small icon.
        sicon = sicon or icon or 'action.svg'
        self.sicon, self.siconBase = iconParts(sicon)
        # If redirect is None, once the transition will be triggered, Appy will
        # perform an automatic redirect:
        # (a) if you were on some "view" page, Appy will redirect you to this
        #     page (thus refreshing it entirely);
        # (b) if you were in a list of objects, Appy will Ajax-refresh the row
        #     containing the object from which you triggered the transition.
        # Case (b) can be problematic if the transition modifies the list of
        # objects, or if it modifies other elements shown outside this list.
        # If you specify "redirect" being
        # ----------------------------------------------------------------------
        # "page"     | case (a) will always apply;
        # ----------------------------------------------------------------------
        # "referer"  | the entire page will be refreshed with the referer page.
        # ----------------------------------------------------------------------
        self.redirect = redirect
        # When a transition is triggered, the corresponding event is added in
        # the object's history. If p_historizeActionMessage is True, the message
        # returned by self.action (if any) will be appended to this event's
        # comment.
        self.historizeActionMessage = historizeActionMessage

    def init(self, workflow, name):
        '''Lazy initialisation'''
        self.workflow = workflow
        self.name = name
        self.labelId = '%s_%s' % (workflow.name, name)

    def __repr__(self):
        return '<transition %s::%s>' % (self.workflow.name, self.name)

    def standardiseStates(self, states):
        '''Get p_states as a list or a list of lists. Indeed, the user may also
           specify p_states as a tuple or tuple of tuples. Having lists allows
           us to easily perform changes in states if required.'''
        if isinstance(states[0], State):
            if isinstance(states, tuple): return list(states)
            return states
        return [[start, end] for start, end in states]

    def getEndStateName(self, wf, startStateName=None):
        '''Returns the name of p_self's end state. If p_self is a
           multi-transition, the name of a specific p_startStateName can be
           given.'''
        if self.isSingle():
            return self.states[1].getName(wf)
        else:
            for start, end in self.states:
                if not startStateName:
                    return end.getName(wf)
                else:
                    if start.getName(wf) == startStateName:
                        return end.getName(wf)

    def getUsedRoles(self):
        '''self.condition can specify a role'''
        res = []
        if isinstance(self.condition, Role):
            res.append(self.condition)
        return res

    def isSingle(self):
        '''If this transition is only defined between 2 states, returns True.
           Else, returns False.'''
        return isinstance(self.states[0], State)

    def addAction(self, action):
        '''Adds an p_action in self.action'''
        actions = self.action
        if not actions:
            self.action = action
        elif isinstance(actions, list):
            actions.append(action)
        elif isinstance(actions, tuple):
            self.action = list(actions)
            self.action.append(action)
        else: # A single action is already defined
            self.action = [actions, action]

    def _replaceStateIn(self, oldState, newState, states):
        '''Replace p_oldState by p_newState in p_states.'''
        if oldState not in states: return
        i = states.index(oldState)
        del states[i]
        states.insert(i, newState)

    def replaceState(self, oldState, newState):
        '''Replace p_oldState by p_newState in self.states.'''
        if self.isSingle():
            self._replaceStateIn(oldState, newState, self.states)
        else:
            for i in range(len(self.states)):
                self._replaceStateIn(oldState, newState, self.states[i])

    def removeState(self, state):
        '''For a multi-state transition, this method removes every state pair
           containing p_state.'''
        if self.isSingle():
            raise WorkflowException('To use for multi-transitions only')
        i = len(self.states) - 1
        while i >= 0:
            if state in self.states[i]:
                del self.states[i]
            i -= 1
        # This transition may become a single-state-pair transition.
        if len(self.states) == 1:
            self.states = self.states[0]

    def setState(self, state):
        '''Configure this transition as being an auto-transition on p_state.
           This can be useful if, when changing a workflow, one wants to remove
           a state by isolating him from the rest of the state diagram and
           disable some transitions by making them auto-transitions of this
           disabled state.'''
        self.states = [state, state]

    def isShowable(self, o):
        '''Is this transition showable ?'''
        return self.show(self.workflow, o) if callable(self.show) else self.show

    def hasState(self, state, isFrom):
        '''If p_isFrom is True, this method returns True if p_state is a
           starting state for p_self. If p_isFrom is False, this method returns
           True if p_state is an ending state for p_self.'''
        stateIndex = 1
        if isFrom:
            stateIndex = 0
        if self.isSingle():
            r = state == self.states[stateIndex]
        else:
            r = False
            for states in self.states:
                if states[stateIndex] == state:
                    r = True
                    break
        return r

    def replaceRoleInCondition(self, old, new):
        '''When self.condition is a tuple or list, this method replaces role
           p_old by p_new. p_old and p_new can be strings or Role instances.'''
        condition = self.condition
        if isinstance(old, Role): old = old.name
        # Ensure we have a list
        if isinstance(condition, tuple): condition = list(condition)
        if not isinstance(condition, list):
            raise WorkflowException('m_replaceRoleInCondition can only be ' \
              'used if transition.condition is a sequence.')
        # Find the p_old role
        i = -1
        found = False
        for cond in condition:
            i += 1
            if isinstance(cond, Role): cond = cond.name
            if cond == old:
                found = True
                break
        if not found: return
        del condition[i]
        condition.insert(i, new)
        self.condition = tuple(condition)

    def isTriggerable(self, o, secure=True):
        '''Can this transition be triggered on p_o ?'''
        # Checks that the current state of the object is a start state for this
        # transition.
        state = o.state
        if self.isSingle():
            if state != self.states[0].name: return
        else:
            found = False
            for start, end in self.states:
                if start.name == state:
                    found = True
                    break
            if not found: return
        # Check that the condition is met, excepted if secure is False
        if not secure: return True
        user = o.user
        # We will need the workflow's prototypical instance
        proto = self.workflow.proto
        if isinstance(self.condition, Role):
            # Condition is a role. Transition may be triggered if the user has
            # this role.
            return user.hasRole(self.condition.name, o)
        elif callable(self.condition):
            return self.condition(proto, o)
        elif type(self.condition) in (tuple, list):
            # It is a list of roles and/or functions. Transition may be
            # triggered if user has at least one of those roles and if all
            # functions return True.
            hasRole = None
            for condition in self.condition:
                # "Unwrap" role names from Role instances
                if isinstance(condition, Role): condition = condition.name
                if isinstance(condition, str): # It is a role
                    if hasRole is None:
                        hasRole = False
                    if user.hasRole(condition, o):
                        hasRole = True
                else: # It is a method
                    r = condition(proto, o)
                    if not r: return r # False or a No instance
            if hasRole != False:
                return True
        else:
            return bool(self.condition)

    def executeAction(self, o):
        '''Executes the action related to this transition'''
        msg = ''
        proto = self.workflow.proto
        if type(self.action) in (tuple, list):
            # We need to execute a list of actions
            for act in self.action:
                msgPart = act(proto, o)
                if msgPart: msg += msgPart
        else: # We execute a single action only
            msgPart = self.action(proto, o)
            if msgPart: msg += msgPart
        return msg

    def getTargetState(self, o):
        '''Gets the target state for this transition'''
        # For a single transition, a single possibility
        if self.isSingle(): return self.states[1]
        sourceName = o.state
        for source, target in self.states:
            if source.name == sourceName:
                return target

    def trigger(self, o, comment=None, doAction=True, doHistory=True,
                doSay=True, reindex=True, secure=True, data=None,
                forceTarget=None):
        '''This method triggers this transition on some p_o(bject). If
           p_doAction is False, the action that must normally be executed after
           the transition has been triggered will not be executed. If
           p_doHistory is False, there will be no trace from this transition
           triggering in p_o's history. If p_doSay is False, we consider the
           transition as being triggered programmatically, and no message is
           returned to the user. If p_reindex is False, object reindexing will
           be performed by the caller method. If p_data is specified, it is a
           dict containing custom data that will be integrated into the history
           event.'''
        # Is that the special _init_ transition ?
        name = self.name
        isInit = name == '_init_'
        # "Triggerability" and security checks
        if not isInit and not self.isTriggerable(o, secure=secure):
            raise Transition.Error(UNTRIGGERABLE % (name, o.url))
        # Identify the target state for this transition
        target = forceTarget or self.getTargetState(o)
        # Add the event in the object history
        event = o.history.add('Trigger', target.name, transition=name,
                              comment=comment)
        # Remember the source state, it will be necessary for executing the
        # common action.
        fromState = o.state if not isInit else None
        # Execute the action that is common to all transitions, if defined. It
        # is named "onTrigger" on the workflow class by convention. This common
        # action is executed before the transition-specific action (if any).
        proto = self.workflow.proto
        if doAction and hasattr(proto, 'onTrigger'):
            proto.onTrigger(o, name, fromState)
        # Execute the transition-specific action
        msg = self.executeAction(o) if doAction and self.action else None
        # Append the action message to the history comment when relevant
        if doHistory and msg and self.historizeActionMessage:
            event.completeComment(msg)
        # Reindex the object if required. Not only security-related indexes
        # (Allowed, State) need to be updated here.
        if reindex and not isInit and not o.isTemp(): o.reindex()
        # Return a message to the user if needed
        if not doSay: return
        return msg or o.translate('object_saved')

    def ui(self, o, mayTrigger):
        '''Return the UiTransition instance corresponding to p_self'''
        return UiTransition(self, o, mayTrigger)

    traverse['fire'] = True
    def fire(self, o):
        '''Executed when a user wants to trigger this transition from the UI'''
        # This requires a database commit
        tool = o.tool
        req = o.req
        handler = o.H()
        handler.commit = True
        # Trigger the transition
        msg = self.trigger(o, req.popupComment, reindex=False)
        # Reindex p_o if required
        if not o.isTemp(): o.reindex()
        # If we are called from an Ajax request, simply return msg
        if handler.isAjax(): return msg
        # If we are viewing the object and if the logged user looses the
        # permission to view it, redirect the user to its home page.
        if msg: o.say(msg)
        # Return to o/view, excepted if the object is not viewable anymore
        if o.guard.mayView(o):
            if self.redirect == 'referer':
                back = handler.headers['Referer']
            else:
                back = o.getUrl(nav=req.nav or 'no', page=req.page or 'main',
                                popup=req.popup)
        else:
            back = tool.computeHomePage()
        o.goto(back)

    @staticmethod
    def getBack(workflow, transition):
        '''Returns the name of the transition (in p_workflow) that "cancels" the
           triggering of p_transition and allows to go back to p_transition's
           start state.'''
        # Get the end state(s) of p_transition
        transition = getattr(workflow, transition)
        # Browse all transitions and find the one starting at p_transition's end
        # state and coming back to p_transition's start state.
        for trName, tr in workflow.__dict__.items():
            if not isinstance(tr, Transition) or (tr == transition): continue
            if transition.isSingle():
                if tr.hasState(transition.states[1], True) and \
                   tr.hasState(transition.states[0], False): return trName
            else:
                startOk = False
                endOk = False
                for start, end in transition.states:
                    if (not startOk) and tr.hasState(end, True):
                        startOk = True
                    if (not endOk) and tr.hasState(start, False):
                        endOk = True
                    if startOk and endOk: return trName

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UiTransition:
    '''Represents a widget that displays a transition'''
    px = Px('''
     <x var="label=transition.title;
             mayTrigger=transition.mayTrigger;
             inButtons=layout == 'buttons'">

      <!-- Picto -->
      <a if="not inButtons" class=":'clickable' if mayTrigger else 'help fake'"
         var="back=transition.getBackHook(o, inButtons, q, backHook)"
         name=":transition.name"
         title=":'' if mayTrigger else transition.reason"
         onclick=":'triggerTransition(%s,this.name,%s,%s)' % \
                    (q(formId), q(transition.confirm), back) \
                    if mayTrigger else ''">
       <img src=":url(transition.icon, base=transition.iconBase)"
            class=":picto"/>
       <div style=":'display:%s' % config.ui.pageDisplay">::label</div>
      </a>

      <!-- Real button -->
      <x if="inButtons" var2="css=ui.Button.getCss(label, inButtons)">
       <input if="mayTrigger" type="button" class=":css" value=":label"
              var="back=transition.getBackHook(o, inButtons, q, backHook)"
              name=":transition.name"
              style=":url(transition.sicon, base=transition.siconBase, \
                          bg='18px 18px')"
              onclick=":'triggerTransition(%s,this.name,%s,%s)' % \
                         (q(formId), q(transition.confirm), back)"/>

       <!-- Fake button, explaining why the transition can't be triggered -->
       <input if="not mayTrigger" type="button"
              class=":'%s fake' % css"
              style=":url(transition.icon, base=transition.siconBase, \
                      bg='18px 18px')"
              value=":label" title=":transition.reason"/></x></x>''',

     js='''
      // Function used for triggering a workflow transition
      function triggerTransition(formId, transition, msg, back) {
        var f = document.getElementById(formId);
        f.action = f.dataset.baseurl + '/' + transition + '/fire';
        submitForm(formId, msg, true, back);
      }''')

    # Fields to copy from Transition to UiTransition
    copied = ('name', 'icon', 'iconBase', 'sicon', 'siconBase')

    def __init__(self, transition, o, mayTrigger):
        _ = o.translate
        # The tied p_transition
        self.transition = transition
        # Copy p_transitions fields here
        for name in UiTransition.copied:
            setattr(self, name, getattr(transition, name))
        self.type = 'transition'
        labelId = transition.labelId
        self.title = _(labelId)
        if transition.confirm:
            msg = o.translate('%s_confirm' % labelId, blankOnError=True) or \
                  o.translate('action_confirm')
            self.confirm = msg
        else:
            self.confirm = ''
        # May this transition be triggered via the UI?
        self.mayTrigger = True
        self.reason = ''
        if not mayTrigger:
            self.mayTrigger = False
            self.reason = mayTrigger.msg
        # Required by the UiGroup
        self.colspan = 1

    def getBackHook(self, o, inButtons, q, backHook=None):
        '''If, when the transition has been triggered, we must ajax-refresh some
           part of the page, this method will return the ID of the corresponding
           DOM node. Else (ie, the entire page needs to be refreshed), it
           returns None.'''
        if inButtons and not self.transition.redirect:
            r = q(backHook or o.iid)
        else:
            r = 'null'
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# ------------------------------------------------------------------------------
from appy.px import Px
from appy.fields.string import String
from appy.gen import WorkflowOwner, Layouts, Show
from appy.gen.wrappers import AbstractWrapper
from appy.gen import utils as gutils
from appy.shared import utils as sutils

# Messages ---------------------------------------------------------------------
SSO_NO_PASSWORD = 'No password is stored for a SSO user.'

# ------------------------------------------------------------------------------
class UserWrapper(AbstractWrapper):
    workflow = WorkflowOwner
    # Users that are automatically logged in
    noLoginUsers = ('system', 'anon')
    specialUsers = noLoginUsers + ('admin',)
    layouts = Layouts.Page.summary

    # Display, in the user strip, links to the User instance of the logged user
    pxUserLink = Px('''
     <span class="userStripText">
      <a if="cfg.userLink" href=":user.url"><img src=":url('user')"/>
       <span style="padding: 0 3px">:user.getTitle()</span></a>
      <x if="not cfg.userLink">:user.getTitle()</x>
      <x var="ctx=cfg.authContext" if="ctx">:ctx.pxLogged</x>
     </span>''')

    def isSpecial(self, includeAdmin=True):
        '''Is this user a Appy-predefined user ?'''
        if includeAdmin:
            return self.login in self.specialUsers
        else:
            return self.login in self.noLoginUsers

    def isAnon(self):
        '''Is the logged user anonymous ?'''
        return self.login == 'anon'

    def getTitle(self, normalized=False, useName=True, nameFirst=False,
                 nameTransform=None, firstNameTransform=None):
        '''Returns a nice name for this user, based on available information:
           "first name"/"name" (if p_useName is True) or title or login. If
           p_normalized is True, special chars (like accents) are converted to
           ascii chars. When p_useName is True, if p_nameFirst is True, the
           result will be "name" / "first name", else it will be
           "first name" / "name".'''
        # A specified transform ("upper", "capitalize", etc) can be specified in
        # p_nameTransform and/or p_firstNameTransform.
        login = self.login
        res = None
        if useName:
            firstName = self.firstName
            name = self.name
            if firstName and name:
                # Apply a transform if requested
                if nameTransform: exec 'name = name.%s()' % nameTransform
                if firstNameTransform:
                    exec 'firstName = firstName.%s()' % firstNameTransform
                # Concatenate first and last names in the right order
                if nameFirst: res = '%s %s' % (name, firstName)
                else: res = '%s %s' % (firstName, name)
        if not res: res = self.title or login
        if not normalized: return res
        return sutils.normalizeString(res)

    def getFirstNameInitial(self, suffix='.', sep='-'):
        '''Gets the first name initial. Take care of composed first names.'''
        r = self.firstName
        if not r: return '?'
        # Work with unicode
        if isinstance(r, str): r = r.decode('utf-8')        
        if '-' in r:
            a, b = r.split('-', 1)
            r = u'%s%s%s%s%s' % (a[0], suffix, sep, b[0], suffix)
        else:
            r = u'%s%s' % (r[0], suffix)
        return r.encode('utf-8')

    def showLogin(self):
        '''When must we show the login field?'''
        if self.o.isTemporary(): return 'edit'
        # The manager has the possibility to change the login itself (local
        # users only).
        if self.user.has_role('Manager') and (self.source == 'zodb'):
            return True
        return Show.E_

    def showName(self):
        '''Name and first name, by default, can not be edited for non-local
           users.'''
        if (self.source != 'zodb'): return Show.E_
        return True

    def showEmail(self):
        '''In most cases, email is the login. Show the field only if it is not
           the case.'''
        # Is this user local ?
        isLocal = self.source == 'zodb'
        # Show the field nevertheless if it is not empty
        if isLocal and not self.isEmpty('email'): return True
        # Hide it if the login is an email
        login = self.login
        if login and String.EMAIL.match(login): return
        # Display the email (read-only if from an external source)
        if not isLocal: return Show.E_
        return True

    def showRoles(self):
        '''Global roles are not visible or editable under all circumstances'''
        # Local users can't be granted global roles
        if self.container != self.tool: return
        # Only a Manager is allowed to grant global roles
        user = self.user
        if user.hasRole('Manager'): return True
        # The user itself, owner of himself, can consult his roles
        if user.hasRole('Owner', self): return Show.E_

    def showResetPassword(self):
        '''Action "reset password" is available to anyone having write access to
           the user, excepted the user himself.'''
        if (self.source == 'zodb') and self.o.mayEdit() and \
           (self.user != self) and not self.isSpecial(includeAdmin=False):
            return 'buttons'

    def validateLogin(self, login):
        '''Is this p_login valid ?'''
        # 2 cases: (1) The user is being created and has no login yet, or
        #          (2) The user is being edited and has already a login, that
        #              can potentially be changed.
        if not self.login or (login != self.login):
            # A new p_login is requested. Check if it is valid and free.
            # Some logins are not allowed.
            if login in self.specialUsers:
                return self.translate('login_reserved')
            # Check that no user or group already uses this login.
            if self.count('User', noSecurity=True, login=login) or \
               self.count('Group', noSecurity=True, login=login):
                return self.translate('login_in_use')
        return True

    def validatePassword(self, password):
        '''Is this p_password valid?'''
        # Password must be at least 8 chars length
        if len(password) < 8:
            return self.translate('password_too_short', mapping={'nb': 8})
        # Ensure the password is not the same as the currently stored password
        # for this user, if any.
        zopeUser = self.getZopeUser()
        if zopeUser:
            from AccessControl.AuthEncoding import pw_validate
            if pw_validate(zopeUser.__, password):
                return self.translate('same_password')
        return True

    def showPassword12(self):
        '''Fields "password1" and "password2", used only for modifying a
           password on an existing user (on specific page "passwords"), must be
           shown only for the user that can modify his password.'''
        # On creation, fields "password3" and "password4" are used
        if self.o.isTemporary(): return
        # When the user itself (we don't check role Owner because a Manager can
        # also own a User instance) wants to edit information about himself.
        if (self.user.login == self.login) and (self.source == 'zodb'):
            return 'edit'

    def showPassword34(self):
        '''Fields "password3" and "password4", used only on the main page for
           entering a password when creating a new user, must be shown only when
           the object is temporary.'''
        if self.o.isTemporary(): return 'edit'

    def showCancel12(self):
        '''Do not show the "cancel" button on page "passwords" if the user is
           forced to change its password.'''
        return not self.changePasswordAtNextLogin

    def encryptPassword(self, clearPassword):
        '''Returns p_clearPassword, encrypted'''
        return self.o.getTool().acl_users._encryptPassword(clearPassword)

    def setPassword(self, newPassword=None, log=True, maxLength=9):
        '''Sets a p_newPassword for self. If p_newPassword is not given, we
           generate one (of p_maxLength chars). This method returns the
           generated password (or simply p_newPassword if no generation
           occurred).'''
        if newPassword is not None:
            msgPart = 'changed'
        else:
            # Take any password field for generating a new password
            fp2 = self.getField('password2')
            newPassword = fp2.generatePassword(maxLength=maxLength)
            msgPart = 'generated'
        login = self.login
        zopeUser = self.getZopeUser()
        tool = self.tool.o
        zopeUser.__ = self.encryptPassword(newPassword)
        req = getattr(tool, 'REQUEST', None)
        if req and hasattr(req, 'userLogin') and (req.userLogin == login):
            # The user for which we change the password is the currently logged
            # user. So update the authentication cookie, too.
            ctx = getattr(req, 'authContext', None)
            gutils.writeCookie(login, newPassword, ctx, self.request)
        if log:
            self.log('password %s for %s.' % (msgPart, login))
        return newPassword

    def doResetPassword(self, noSecurity=False):
        '''Triggered from the ui, this method defines a new, automatically
           generated password for this user and returns it in ui.'''
        if not noSecurity and not self.user.allows('write'):
            self.raiseUnauthorized()
        password = self.setPassword()
        self.changePasswordAtNextLogin = True
        self.say('New password for this user is %s.' % password)

    def setEncryptedPassword(self, encryptedPassword):
        '''Sets p_encryptedPassword for this user. m_setPassword above starts
           for a clear (given or generated) password. This one simply sets an
           already encrypted password for this user.'''
        self.getZopeUser().__ = encryptedPassword

    def getEncryptedPassword(self):
        '''Gets the encrypted password for this user'''
        if self.source == 'sso': raise Exception(SSO_NO_PASSWORD)
        return self.getZopeUser()._getPassword()

    def showEncryptedPassword(self):
        '''Allow to show encrypted password to admins only, via the xml
           layout.'''
        if self.user.hasRole('Manager') and (self.source != 'sso'): return 'xml'

    def checkPassword(self, clearPassword):
        '''Returns True if p_clearPassword is the correct password for this
           user.'''
        encryptedPassword = self.getZopeUser()._getPassword()
        from AccessControl.AuthEncoding import pw_validate
        return pw_validate(encryptedPassword, clearPassword)

    def getMailRecipient(self):
        '''Returns, for this user, the "recipient string" (first name, name,
           email) as can be used for sending an email.'''
        r = self.email or self.login
        # Ensure this is really an email
        if not String.EMAIL.match(r): return
        return '%s <%s>' % (self.getTitle(), r)

    def setLogin(self, oldLogin, newLogin, reindex=True):
        '''Changes the login of this user from p_oldLogin to p_newLogin'''
        self.login = newLogin
        # Update the corresponding Zope-level user
        aclUsers = self.o.acl_users
        zopeUser = aclUsers.data[oldLogin]
        zopeUser.name = newLogin
        del aclUsers.data[oldLogin]
        aclUsers.data[newLogin] = zopeUser
        # Update the email if the email corresponds to the login
        email = self.email
        if email == oldLogin:
            self.email = newLogin
        # Update the title
        self.updateTitle()
        # Browse all objects of the database and update potential local roles
        # that referred to the old login.
        context = {'nb': 0, 'total': 0, 'old': oldLogin, 'new': newLogin}
        for className in self.tool.o.getAllClassNames():
            self.compute(className, context=context, noSecurity=True,
                         expression="ctx['nb'] += obj.o.applyUserIdChange(" \
                                   "ctx['old'], ctx['new']); ctx['total'] += 1")
        self.log('login "%s" renamed to "%s".' % (oldLogin, newLogin))
        self.log('login change: local roles and/or history updated in %d/%d ' \
                 'object(s).' % (context['nb'], context['total']))
        # Reindex p_self, excepted if p_reindex is False
        if reindex: self.reindex()

    def getGrantableRoles(self):
        '''Returns the list of roles that the admin can grant to a user'''
        res = []
        for role in self.o.getProductConfig().grantableRoles:
            res.append( (role, self.translate('role_%s' % role)) )
        return res

    def validate(self, new, errors):
        '''Inter-field validation'''
        self.o._oldLogin = None
        # Check that passwords match, either on page "passwords" (passwords 1
        # and 2) or on page "main" (passwords 3 and 4).
        for i in (1, 3):
            passwordA = 'password%d' % i
            passwordB = 'password%d' % (i + 1)
            if hasattr(new, passwordA) and \
               (getattr(new, passwordA) != getattr(new, passwordB)):
                msg = self.translate('passwords_mismatch')
                setattr(errors, passwordA, msg)
                setattr(errors, passwordB, msg)
            # Remember the previous login
            if self.login: self.o._oldLogin = self.login
        return self._callCustom('validate', new, errors)

    def updateTitle(self):
        '''Sets a title for this user'''
        self.title = self.getTitle(nameFirst=True)

    def ensureAdminIsManager(self):
        '''User "admin" must always have role "Manager"'''
        if self.o.id == 'admin':
            roles = self.roles
            if 'Manager' not in roles:
                if not roles: roles = ['Manager']
                else: roles.append('Manager')
                self.roles = roles

    def getSupTitle(self, nav):
        '''Display a specific icon if the user is a local copy of an external
           (ldap/sso) user.'''
        res = self._callCustom('getSupTitle', nav) or ''
        if self.source == 'zodb': return res
        # For external users note the source and last synchronization date in
        # the icon's title.
        tool = self.o.getTool()
        if self.isEmpty('syncDate'):
            suffix = ''
        else:
            suffix = ' - sync@ %s' % tool.formatDate(self.syncDate)
        return '<img src="%s/ui/external.png" title="%s%s" class="help"/>%s' % \
               (tool.getSiteUrl(), self.source.upper(), suffix, res)

    def getZopeUser(self):
        '''Gets the Zope user corresponding to this user'''
        if self.source == 'zodb':
            return self.o.acl_users.data.get(self.login, None)
        return self.o._zopeUser

    def getLogins(self, groupsOnly=False):
        '''Gets all the logins that can "match" this user: it own login
           (excepted if p_groupsOnly is True) and the logins of all the groups
           he belongs to.'''
        # Try first to get those logins from a cache on the request, if this
        # user corresponds to the logged user.
        rq = self.request
        if hasattr(rq, 'userLogins') and (rq.userLogin == self.login):
            return rq.userLogins
        # Compute it
        res = [group.login for group in self.groups]
        if not groupsOnly: res.append(self.login)
        return res

    def getRoles(self):
        '''This method returns all the global roles for this user, not simply
           self.roles, but also "ungrantable roles" (like Anonymous or
           Authenticated) and roles inherited from group membership.'''
        # Try first to get those roles from a cache on the request, if this user
        # corresponds to the logged user.
        rq = self.request
        if hasattr(rq, 'userRoles') and (rq.userLogin == self.login):
            return rq.userRoles
        # Compute it
        res = list(self.roles)
        # Add ungrantable roles
        res.append(self.isAnon() and 'Anonymous' or 'Authenticated')
        # Add group global roles
        for group in self.groups:
            for role in group.roles:
                if role not in res: res.append(role)
        return res

    def getRolesFor(self, obj):
        '''Gets the roles the user has in the context of p_obj: its global roles
           + its roles which are local to p_obj.'''
        obj = obj.o
        # Start with user global roles
        res = self.getRoles()
        # Add local roles, granted to the user directly or to one of its groups
        localRoles = obj.appy().localRoles
        if not localRoles: return res
        # Gets the logins of this user and all its groups. Create a new list.
        # Indeed, "res" can be the actual list of global roles as cached in the
        # request or the currently logged user.
        res = list(res)
        logins = self.getLogins()
        for login, roles in localRoles.iteritems():
            # Ignore logins not corresponding to this user
            if login not in logins: continue
            for role in roles:
                if role not in res: res.append(role)
        return res

    def addRole(self, role):
        '''Adds p_role to the user's global roles'''
        roles = self.roles
        if role in roles: return
        roles = list(roles)
        roles.append(role)
        self.roles = roles
        self.getZopeUser().roles = roles

    def delRole(self, role):
        '''Remove p_role from the user's global roles'''
        roles = self.roles
        if role not in roles: return
        roles = list(roles)
        roles.remove(role)
        self.roles = roles
        self.getZopeUser().roles = roles

    def hasRole(self, role, obj=None):
        '''Has the logged user some p_role? If p_obj is None, check if the user
           has p_role globally; else, check if he has this p_role in the context
           of p_obj.

           p_role can also be a list/tuple of roles. In this case, the method
           returns True if the user has at least one of the listed roles.'''
        # Try first with the user's global roles, excepted if p_obj is in
        # "local" mode.
        noo = obj is None
        if noo or not obj.o.getLocalRolesOnly():
            r = sutils.stringIsAmong(role, self.getRoles())
            if noo or r: return r
        # Check now p_obj's local roles
        localRoles = obj.appy().localRoles
        if not localRoles: return
        logins = self.getLogins()
        for login, roles in localRoles.iteritems():
            if (login in logins) and sutils.stringIsAmong(role, roles):
                return True
    has_role = hasRole  # "has_role" will be removed in Appy 1.0

    def hasPermission(self, permission, obj):
        '''Has the logged user p_permission on p_obj ?'''
        obj = obj.o
        # What are the roles which are granted p_permission on p_obj?
        allowedRoles = obj.getRolesFor(permission)
        if not allowedRoles: return
        # Grant access based on global user roles (that include ungrantable
        # roles like Authenticated or Anonymous), excepted if p_obj is in
        # "local" mode.
        if not obj.getLocalRolesOnly():
            for role in self.getRoles():
                if role in allowedRoles: return True
        # Grant access based on local roles
        localRoles = obj.appy().localRoles
        if not localRoles: return
        # Gets the logins of this user and all its groups
        userLogins = self.getLogins()
        for login, roles in localRoles.iteritems():
            # Ignore logins not corresponding to this user
            if login not in userLogins: continue
            for role in roles:
                if role in allowedRoles: return True
    has_permission = hasPermission #"has_permission" will be removed in Appy 1.0

    def onEdit(self, created):
        '''Triggered when a User is created or updated'''
        login = self.login
        # Refresh the title, computed from other fields
        self.updateTitle()
        # Is it a local User or a LDAP User?
        isLocal = self.source == 'zodb'
        # Ensure correctness of some infos about this user
        if isLocal: self.ensureAdminIsManager()
        if created:
            # Create the corresponding Zope user
            from AccessControl.User import User as ZopeUser
            password = self.encryptPassword(self.password3)
            zopeUser = ZopeUser(login, password, self.roles, ())
            # Add it in acl_users if it is a local user
            if isLocal:
                self.o.acl_users.data[login] = zopeUser
                # If user creation was made by an admin, force the user to
                # change its password at its first login.
                if not self.isSpecial() and self.user.hasRole('Manager'):
                    self.changePasswordAtNextLogin = True
            # Add it in self.o._zopeUser if it is a LDAP or SSO user
            else: self.o._zopeUser = zopeUser
            # Remove our own password copies
            self.password3 = self.password4 = ''
        else:
            # Update the login itself if the user has changed it
            if hasattr(self.o.aq_base, '_oldLogin'):
                oldLogin = self.o._oldLogin
                if oldLogin and (oldLogin != login):
                    self.setLogin(oldLogin, login, reindex=False)
                del self.o._oldLogin
            # Update roles at the Zope level
            zopeUser = self.getZopeUser()
            zopeUser.roles = self.roles
            # Update the password if the user has entered new ones
            rq = self.request
            if rq.get('page', 'main') == 'passwords':
                self.setPassword(rq['password1'])
                self.password1 = self.password2 = ''
                self.changePasswordAtNextLogin = False
        # "self" must be owned by its Zope user
        self.addLocalRole(login, 'Owner')
        # If the user was created by anon|system, anon|system can't stay Owner
        self.deleteLocalRole(('anon', 'system'))
        return self._callCustom('onEdit', created)

    def convertTo(self, source):
        '''Convert a user as if, now, he would be of p_source, that is different
           from self.source.'''
        # p_self.source and p_source must be different
        origSource = self.source
        if origSource == source: return
        zopeUser = self.getZopeUser()
        # Perform the conversion
        if origSource == 'zodb':
            # Convert a local user to an external user.
            # Transfer the Zope user from acl_users to p_self.
            del self.o.acl_users.data[self.login]
            self.o._zopeUser = zopeUser
        else:
            # The user is external
            if source == 'zodb':
                # Convert an external user to a local user.
                # Transfer the Zope user from p_self to acl_users.
                self.o.acl_users.data[self.login] = zopeUser
                del self.o._zopeUser
        # Update p_self's parameters
        self.source = source
        from DateTime import DateTime
        self.syncDate = DateTime()
        # This flag is not relevant for external users
        if self.source != 'zodb':
            self.changePasswordAtNextLogin = False
        # Log the operation
        self.log('user %s converted from %s to %s.' % \
                 (self.login, origSource, source))

    def mayEdit(self):
        '''No one can edit users "system" and "anon"'''
        if self.o.id in ('anon', 'system'): return
        # Call custom "mayEdit" when present
        custom = self._getCustomMethod('mayEdit')
        if custom: return self._callCustom('mayEdit')
        return True

    def onDelete(self):
        '''Before deleting myself, I must delete the corresponding Zope user
           (for local users only).'''
        if self.source == 'zodb': del self.o.acl_users.data[self.login]
        self.log('user %s deleted.' % self.login)
        # Call a custom "onDelete" if any
        return self._callCustom('onDelete')

    def mayDelete(self):
        '''No one can delete special users'''
        if self.isSpecial(): return
        # Call custom "mayDelete" when present
        custom = self._getCustomMethod('mayDelete')
        if custom: return self._callCustom('mayDelete')
        return True
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.data import rtlLanguages

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Languages:
    '''Manages user language-related elements'''

    @classmethod
    def getDirection(self, lang):
        '''Determines if p_lang is a LTR (left-to-right) or RTL (right-to-left)
           language.'''
        # It returns a 3-tuple  (dir, dleft, dright)
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # dir    | String "ltr" for a left-to-right language, "rtl" for a
        #        | right-to-left language;
        # dleft  | String "left" for a LTR and "right" for a RTL language
        # dright | String "right" for a LTR and "left" for a RTL language
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        isRtl = lang in rtlLanguages
        return ('rtl', 'right', 'left') if isRtl else ('ltr', 'left', 'right')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Deployment system for Appy sites and apps'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os, sys
from pathlib import Path
from appy.model.utils import Object as O
from appy.deploy.subversion import Subversion

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def parseSpecifier(specifier):
    '''Parses a string specifier of the form <attr>=<value>[,<attr>=<value>]'''
    r = O()
    for part in specifier.split(','):
        name, value = part.split('=', 1)
        r[name] = value
    return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Target:
    '''Represents a app deployed on a site on a distant machine'''

    def __init__(self, sshHost, sshPort=22, sshLogin='root', sshKey=None,
                 sitePath=None, sitePort=8000, siteApp=None,
                 siteOwner='appy:appy', siteDependencies=None):
        # The name of the distant host, for establishing a SSH connection
        self.sshHost = sshHost
        # The port for the SSH connection
        self.sshPort = sshPort
        # The login used to connect to the host in SSH
        self.sshLogin = sshLogin
        # The private key used to connect to the host in SSH
        self.sshKey = sshKey
        # Information about the Appy site on the target
        # ~~~
        # The path to the site. Typically: /home/appy/<siteName>
        self.sitePath = sitePath
        # The port on which this site will listen
        self.sitePort = sitePort
        # The specifier of the distant app. For example:
        #    url=https://svn.forge.pallavi.be/appy-python-3/trunk,name=appy
        self.siteApp = siteApp
        # The owner of the distant site. Typically: appy:appy.
        self.siteOwner = siteOwner
        # A list of Python dependencies to install on the distant app, in its
        # "lib" folder. For every dependency, a specifier must be used, with the
        # same syntax as attribute "siteApp" hereabove.
        self.siteDependencies = siteDependencies or []

    def __repr__(self):
        '''p_self's string representation'''
        return '<Target %s:%d>' % (self.sshHost, self.sshPort)

    def execute(self, command):
        '''Executes p_command on this target'''
        r = ['ssh', '%s@%s' % (self.sshLogin, self.sshHost), '"%s"' % command]
        # Determine port
        if self.sshPort != 22: r.insert(1, '-p%d' % self.sshPort)
        # Determine "-i" option (path to the private key)
        if self.sshKey: r.insert(1, '-i %s' % self.sshKey)
        # Build the complete command
        r = ' '.join(r)
        print('Executing: %s...' % r)
        os.system(r)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Deployment configuration'''

    def __init__(self):
        # This dict stores all the known targets for deploying this app. Keys
        # are target names, values are Target instances. The default target must
        # be defined at key "default".
        self.targets = {}

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
NO_CONFIG = 'The "deploy" config was not found in config.deploy.'
NO_TARGET = 'No target was found on config.deploy.targets.'
TARGET_KO = 'Target "%s" not found. Available target(s): %s.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Deployer:
    '''App deployer'''

    # apt command for installing packages non interactively
    apt = 'DEBIAN_FRONTEND=noninteractive apt-get -yq install'

    # OS packages being Appy dependencies
    osDependencies = 'libreoffice subversion python3-pip apache2 imagemagick'

    def __init__(self, appPath, sitePath, command, targetName='default'):
        # The path to the app
        self.appPath = appPath
        # The path to the reference, local site, containing targets definition
        self.sitePath = sitePath
        # The chosen target (name)
        self.targetName = targetName or 'default'
        self.target = None # Will hold a Target instance
        # The command to execute
        self.command = command
        # The app config
        self.config = None

    def quote(self, arg):
        '''Surround p_arg with quotes'''
        r = arg if isinstance(arg, str) else str(arg)
        return "'%s'" % r

    def buildPython(self, statements):
        '''Builds a p_command made of these Python p_statements'''
        return "python3 -c \\\"%s\\\"" % ';'.join(statements)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                              Commands
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Command for consulting the last lines in a target's app.log file
    tail = 'tail -f -n %d %s/var/app.log'

    def list(self):
        '''Lists the available targets on the config'''
        infos = []
        for name, target in self.config.deploy.targets.items():
            info = '*** Target: %s\n%s' % (name, target)
            infos.append(info)
        infos = '\n\n'.join(infos)
        print('Available target(s) for app "%s", from reference site "%s":\n' \
              '%s' % (self.appPath.name, self.sitePath.name, infos))

    def info(self):
        '''Retrieve info about the target OS'''
        self.target.execute('cat /etc/lsb-release')

    def install(self):
        '''Installs required dependencies on the target via "apt" and "pip3" and
           create special user "appy" on the server.'''
        target = self.target
        commands = [
          # Install required dependencies via Aptitude
          '%s %s' % (self.apt, self.osDependencies),
          # Install Appy and dependencies via pip
          'pip3 install appy',
          # Create special user "appy"
          'adduser --disabled-password --gecos \'\' appy'
        ]
        target.execute(';'.join(commands))

    def site(self):
        '''Creates an Appy site on the distant server'''
        t = self.target
        # Build args to appy/bin/make
        q = self.quote
        args = [q(t.sitePath), q('-a'), q(t.siteApp), q('-p'), q(t.sitePort),
                q('-o'), q(t.siteOwner)]
        if t.siteDependencies:
            args += [q('-d'), q(' '.join(t.siteDependencies))]
        # Build the statements to pass to the distant Python interpreter
        statements = [
          'import sys', 'from appy.bin.make import Make',
          "sys.argv=['make.py','site',%s]" % ','.join(args),
          'Make().run()'
        ]
        command = self.buildPython(statements)
        # Execute it
        t.execute(command)

    def update(self):
        '''Performs an update of all software known to the site and coming from
           external sources (app and dependencies) and (re)starts the site.'''
        target = self.target
        # Build the commands to update the app and dependencies
        commands = []
        svn = Subversion(parseSpecifier(target.siteApp))
        lib = Path(target.sitePath) / 'lib'
        svnUp, local = svn.build('update', lib)
        commands.append(svnUp)
        for dep in target.siteDependencies:
            svn = Subversion(parseSpecifier(dep))
            svnUp, local = svn.build('update', lib)
            commands.append(svnUp)
        # Build the command to restart the distant site
        restart = '%s/bin/site restart' % target.sitePath
        commands.append(restart)
        commands.append(self.tail % (100, target.sitePath))
        # The commands will be ran with target.siteOwner
        owner = target.siteOwner.split(':')[0]
        command = "su %s -c '%s'" % (owner, ';'.join(commands))
        target.execute(command)

    def view(self):
        '''Launch a command "tail -f" on the target's app.log file'''
        target = self.target
        target.execute(self.tail % (200, target.sitePath))

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                             Main method
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def run(self):
        '''Performs p_self.command on the specified p_self.targetName'''
        # Add the relevant paths to sys.path
        for path in (self.sitePath, self.sitePath / 'lib', self.appPath.parent):
            sys.path.insert(0, str(path))
        # Get the config and ensure it is complete
        self.config = __import__(self.appPath.name).Config
        cfg = self.config.deploy
        if not cfg:
            print(NO_CONFIG)
            sys.exit(1)
        targets = cfg.targets
        if not targets:
            print(NO_TARGET)
            sys.exit(1)
        # Get the target
        name = self.targetName
        target = self.target = targets.get(name)
        if not target:
            print(TARGET_KO % (name, ', '.join(targets)))
            sys.exit(1)
        getattr(self, self.command)()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This module defines classes allowing to store an object's history'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Most of an object's history is related to its workflow. This is why this
# module lies within appy.model.workflow. That being said, object history also
# stores not-workflow-based events like data changes.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import persistent
from DateTime import DateTime
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping

from appy.px import Px
from appy.model.batch import Batch
from appy.xml.escape import Escape
from appy.utils import string as sutils

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class EventIterator:
    '''Iterator for history events'''

    def __init__(self, history, eventType=None, condition=None,
                 context=None, chronological=False, i=None):
        # The history containing the events to walk
        self.history = history
        # The types of events to walk. None means that all types are walked.
        # When specified, p_eventType must be the name of a concrete Event
        # class, or a list/tuple of such class names.
        self.eventType = (eventType,) if isinstance(eventType, str) \
                                      else eventType
        # An additional condition, as a Python expression (getting "event" in
        # its context) that will dismiss the event if evaluated to False.
        self.condition = condition
        # A context that will be given to the condition
        self.context = context
        # If chronological is True, events are walked in chronological order.
        # Else, they are walked in their standard, anti-chronological order.
        self.chronological = chronological
        # The index of the currently walked event
        if i is None:
            self.i = len(history) - 1 if chronological else 0
        else:
            self.i = i

    def increment(self):
        '''Increment p_self.i, or decrement it if we walk events in
           chronological order.'''
        if self.chronological:
            self.i -= 1
        else:
            self.i += 1

    def typeMatches(self, event):
        '''Has p_event the correct type according to p_self.eventType ?'''
        # If no event type is defined, p_event matches
        if self.eventType is None: return True
        return event.__class__.__name__ in self.eventType

    def conditionMatches(self, event):
        '''Does p_event matches p_self.condition ?'''
        # If no condition is defined, p_event matches
        if self.condition is None: return True
        # Update the evaluation context when appropriate
        if self.context:
            locals().update(context)
        return eval(self.condition)

    def __iter__(self): return self
    def __next__(self):
        '''Return the next matching event'''
        try:
            event = self.history[self.i]
        except IndexError:
            # There are no more events, we have walked them all
            raise StopIteration
        # Does this event match ?
        if self.typeMatches(event) and self.conditionMatches(event):
            # Yes
            self.increment()
            return event
        else:
            # Try to return the next element
            self.increment()
            return self.__next__()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Event(persistent.Persistent):
    '''An object's history is made of events'''
    # A format for representing dates at various (non ui) places
    dateFormat = '%Y/%m/%d %H:%M'

    # Some events can be deleted from the ui, but most aren't
    deletable = False

    # The PX displaying details about an event (basically the event's comment)
    pxDetail = Px('''
     <x var="eventId=event.getId()">
      <img if="event.mayEditComment(config, user, isManager)"
           class="clickable iconS" style="float:right" src=":url('edit.svg')"
           title=":_('object_edit')"
           onclick=":'onHistoryEvent(%s,%s,%s,%s)' % \
                    (q('Edit'), q(o.iid), q(event.date), q(eventId))"/>
      <span id=":eventId">::event.getComment()</span>
     </x>''')

    def __init__(self, login, state, date, comment=None):
        # The login of the user that has triggered the event
        self.login = login
        # The name of the state into which the object is after the event
        # occurred. It means that an object's current state is stored in this
        # attribute, on the most recent event in its history.
        self.state = state
        # When did this event occur ?
        self.date = date
        # A textual optional comment for the event
        self.comment = comment

    def clone(self):
        '''Create an return a clone from myself'''
        params = self.__dict__.copy()
        login = params.pop('login')
        state = params.pop('state')
        date = params.pop('date')
        return self.__class__(login, state, date, **params)

    def completeComment(self, comment, sep='<br/><br/>'):
        '''Appends p_comment to the existing p_self.comment'''
        if not self.comment:
            self.comment = comment
        else:
            self.comment =  '%s%s%s' % (self.comment, sep, comment)

    def getTypeName(self):
        '''Return the class name, possibly completed with sub-class-specific
           information.'''
        return self.__class__.__name__

    def getLabel(self, o):
        '''Returns the i18n label for this event'''
        # To be overridden

    def getComment(self, empty='-'):
        '''Returns the formatted version of p_self.comment'''
        comment = self.comment
        if not comment: return empty
        Escape.xhtml(comment, p=True)

    def mayEditComment(self, config, user, isManager):
        '''May p_user edit the comment on this event ?'''
        if not config.model.editHistoryComments: return
        return isManager or (user.login == self.login)

    def getId(self):
        '''Return an ID for this history p_event, based on its date.'''
        return str(self.date.millis())

    def __repr__(self):
        '''String representation'''
        date = self.date.strftime(Event.dateFormat)
        return '<%s by %s on %s, state %s>' % \
               (self.getTypeName(), self.login, date, self.state)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Trigger(Event):
    '''This event represents a transition being triggered'''

    def __init__(self, login, state, date, **params):
        # Extract the name of the transition from p_params
        self.transition = params.pop('transition')
        Event.__init__(self, login, state, date, **params)

    def getTypeName(self):
        '''Return the class name and the transition name'''
        return 'Trigger %s' % self.transition

    def getLabel(self, o):
        '''Returns the label for the corresponding transition'''
        if self.transition == '_init_':
            r = 'Base_creator'
        else:
            transition = o.getWorkflow().transitions.get(self.transition)
            r = transition.labelId if transition else self.transition
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Action(Event):
    '''This event represents an action (field) being performed'''

    def __init__(self, login, state, date, **params):
        # Extract the name of the action from p_params
        self.action = params.pop('action')
        Event.__init__(self, login, state, date, **params)

    def getTypeName(self):
        '''Return the class name and the transition name.'''
        return 'Action %s' % self.action

    def getLabel(self, o):
        '''Returns the label for the corresponding action'''
        return o.getField(self.action).labelId

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Change(Event):
    '''This event represents a data change'''

    # Due to the mess sometimes occurring within XHTML diffs, it may be
    # preferable to forget about the change by allowing the user to delete the
    # corresponding event.
    deletable = True

    # Types of entries in a diff
    diffEntries = ('insert', 'delete')

    # Showing details about a change = displaying the previous values of the
    # fields whose values were modified in this change.
    pxDetail = Px('''
     <table class="changes" width="100%" if="event.changes">
      <tr>
       <th align=":dleft">:_('modified_field')</th>
       <th align=":dleft">:_('previous_value')</th>
      </tr>
      <tr for="name, value in event.changes.items()" valign="top"
          var2="fname=name if isinstance(name, str) else name[0];
                lg=None if isinstance(name, str) else name[1];
                field=o.getField(fname)">
       <td><x>::_(field.labelId) if field else fname</x>
           <x if="lg">:' (%s)' % ui.Language.getName(lg)</x></td>
       <td>::field.getHistoryValue(o, value, history.data.index(event), lg) \
             if field else value</td>
      </tr>
     </table>''')

    def __init__(self, login, state, date, **params):
        # Extract changed fields from p_params. Attribute "changes" stores, in a
        # dict, the previous values of fields whose values have changed. If the
        # dict's key is of type...
        # ----------------------------------------------------------------------
        # str   | it corresponds to the name of the field;
        # tuple | it corresponds to a tuple (s_name, s_language) and stores the
        #       | part of a multilingual field corresponding to s_language.
        # ----------------------------------------------------------------------
        self.changes = PersistentMapping(params.pop('changes'))
        Event.__init__(self, login, state, date, **params)

    def hasField(self, name, language=None):
        '''Is there, within p_self's changes, a change related to a field whose
           name is p_name ?'''
        if name in self.changes: return True
        # Search for a key of the form (name, language)
        for key in self.changes.keys():
            if isinstance(key, tuple) and (key[0] == name) and \
               (not language or (language == key[1])):
                return True

    def getValue(self, name, language=None):
        '''Gets the value as stored on this change, for field name p_name, or
           its p_language part.'''
        key = (name, language) if language else name
        return self.changes.get(key)

    def getLabel(self, o):
        '''Returns the translated label for a data change'''
        return 'event_Change'

    def getDiffTexts(self, o):
        '''Returns a tuple (insertText, deleteText) containing texts to show on,
           respectively, inserted and deleted chunks of text in a XHTML diff.'''
        user = o.search1('User', login=self.login, secure=False)
        mapping = {'userName': user.getTitle() if user else self.login}
        r = []
        tool = o.tool
        for type in self.diffEntries:
            msg = self.translate('history_%s' % type, mapping=mapping)
            date = tool.formatDate(self.date)
            r.append('%s: %s' % (date, msg))
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Custom(Event):
    '''Represents a custom event, for which the label is chosen by the
       developer.'''

    def __init__(self, login, state, date, **params):
        # Extract the label to use to name the event
        self.label = params.pop('label')
        Event.__init__(self, login, state, date, **params)

    def getLabel(self, o):
        '''Returns the translated label for a data change'''
        return self.label

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Link(Event):
    '''Represents a link, via a Ref field, between 2 objects. The event is
       stored in the source object's history.'''

    # Although it may seem similar to a Change, it does not inherit from it.
    # Indeed, the complete list of previously linked objects is not stored:
    # instead, the title of the newly linked object is set as comment in this
    # Link event.

    def __init__(self, login, state, date, **params):
        # We may have linked an existing object or a newly created one
        self.addition = params.pop('addition')
        Event.__init__(self, login, state, date, **params)

    def getLabel(self, o):
        '''Returns the translated label for a link'''
        return 'event_%s' % ('Addition' if self.addition else 'Link')

class Unlink(Event):
    '''Represents an object being unlinked from another one via a Ref field'''

    def __init__(self, login, state, date, **params):
        # We may have simply unlinked an object or it may have been deleted
        self.deletion = params.pop('deletion')
        Event.__init__(self, login, state, date, **params)

    def getLabel(self, o):
        '''Returns the translated label for an unlink'''
        return 'event_%s' % ('Deletion' if self.addition else 'Unlink')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class History(PersistentList):
    '''Object history is implemented as a list, sorted in antichronological
       order, of history events.'''

    # This PX displays history events
    events = Px('''
     <div var="history=o.history;
               batchSize=historyMaxPerPage|req.maxPerPage;
               batch=history.getBatch(o, batchSize);
               isManager=user.hasRole('Manager')"
          if="batch.length" id=":batch.hook">
      <script>:history.getAjaxData(batch)</script>

      <!-- Navigate between history pages -->
      <div align=":dright">:batch.pxNavigate</div>

      <!-- History -->
      <table width="100%" class="history">
       <tr>
        <th align=":dleft">:_('object_action')</th>
        <th align=":dleft">:_('object_author')</th>
        <th align=":dleft">:_('action_date')</th>
        <th align=":dleft">:_('action_comment')</th>
       </tr>
       <tr for="event in batch.objects"
           class=":loop.event.odd and 'even' or 'odd'" valign="top">
        <td><x>:_(event.getLabel(o))</x>
         <img if="isManager and event.deletable" class="clickable iconS"
              src=":url('deleteS.svg')"
              onclick=":'onHistoryEvent(%s,%s,%s)' % \
                        (q('Delete'), q(o.iid), q(event.date))"/></td>
        <td var="creator=o.search1('User', login=event.login)">
         <x>:creator.getTitle() if creator else (event.login or '?')</x></td>
        <td>:tool.formatDate(event.date)</td>
        <td>:event.pxDetail</td>
       </tr>
      </table>
     </div>''')

    view = Px('''
     <div if="not o.isTemp()"
       var2="history=o.history; hasHistory=not history.isEmpty();
             createComment=history[-1].comment">
      <table width="100%" class="header" cellpadding="0" cellspacing="0">
       <tr>
        <td colspan="2" class="by">
         <!-- Creator and last modification date -->
         <x>:_('Base_creator')</x> 
          <x>:user.getTitleFromLogin(o.creator)</x> 

         <!-- Creation and last modification dates -->
         <x>:_('Base_created')</x> 
         <x var="created=o.created; modified=o.modified">
          <x>:tool.Date.format(tool, created, withHour=True)</x>
          <x if="modified != created">&mdash;
           <x>:_('Base_modified')</x>
           <x>:tool.Date.format(tool, modified, withHour=True)</x>
          </x>
         </x>

         <!-- State -->
         <x> &mdash; <x>:_('Base_state')</x> : 
            <b>:_(o.getLabel(o.state, field=False))</b></x>

         <!-- Initial comment -->
         <div if="createComment" class="topSpace">::createComment</div>
        </td>
       </tr>

       <!-- History entries -->
       <tr if="hasHistory"><td colspan="2">:history.events</td></tr>
      </table>
     </div>''',

     css='''
      .header { margin-bottom:5px; background-color:#f3f3f7;
                border:3px solid white }
      .by { padding:7px }
      .history>tbody>tr>td { padding:8px }
      .history>tbody>tr>th { font-style:italic; text-align:left;
                             padding:10px 5px; background-color:white }
      .changes { margin: 4px 0 }
      .changes>tbody>tr>td, .changes>tbody>tr>th { border:1px solid #e0e0e0;
                                                   padding:5px }
     ''')

    def __init__(self, o):
        PersistentList.__init__(self)
        # A reference to the object for which p_self is the history
        self.o = o
        # The last time the object has been modified
        self.modified = None

    def add(self, type, state=None, **params):
        ''''Adds a new event of p_type (=the name one an Event sub-class into
            the history.'''
        # Get the login of the user performing the action
        if 'login' in params:
            login = params.pop('login')
        else:
            login = self.o.user.login
        # Get the date and time of the action
        if 'date' in params:
            date = params.pop('date')
        else:
            date = DateTime()
        # For a trigger event, p_state is the new object state after the
        # transition has been triggered. p_state is a name (string) and not a
        # State instance. The name of the triggering transition must be in
        # p_params, at key "transition".
        # For a change event, no state is given, because there is no state
        # change, but we will copy, on the change event, the state from the
        # previous event. That way, the last event in the history will always
        # store the object's current state.
        state = state or self[0].state
        # Create the event
        event = eval(type)(login, state, date, **params)
        # Insert it at the first place within the anti-chronological list
        self.insert(0, event)
        # Initialise self.modified if still None
        if self.modified is None: self.modified = event.date
        return event

    def iter(self, **kwargs):
        '''Returns an iterator for browsing p_self's events'''
        return EventIterator(self, **kwargs)

    def isEmpty(self, name=None):
        '''Is this history empty ? If p_name is not None, the question becomes:
           has p_self.o an history for field named p_name ?'''
        # An history containing a single entry is considered empty: this is the
        # special _init_ virtual transition representing the object creation.
        if len(self) == 1: return True
        # At this point, the complete history can be considered not empty
        if name is None: return
        # Check if history is available for field named p_name
        empty = True
        for event in self.iter(eventType='Change', \
                               condition="event.hasField('%s')" % name):
            # If we are here, at least one change concerns the field
            empty = False
            break
        return empty

    def getCurrentValues(self, o, fields):
        '''Called before updating p_o, this method remembers, for every
           historized field from p_fields, its current value.'''
        r = {} # ~{s_fieldName: currentValue}~
        # p_fields can be a list of fields or a single field
        fields = fields if isinstance(fields, list) else [fields]
        # Browse fields
        for field in fields:
            if not field.getAttribute(o, 'historized'): continue
            r[field.name] = field.getComparableValue(o)
        return r

    def getNewer(self, field, value, i, language=None):
        '''Find the newer version of this p_field p_value in p_self, starting at
           index p_i, or, if not found, take the value as currently stored on
           p_self.o.'''
        # p_value was found in this history, at index p_i+1
        found = False
        name = field.name
        for event in self.iter(eventType='Change', chronological=True, i=i):
            if event.hasField(name, language=language):
                # A newer version exists in this history
                newer = event.getValue(name, language)
                found = True
                break
        # If no newer version was found in p_o's history, take the value being
        # currently stored on p_o for p_self.
        if not found:
            newer = getattr(self.o, name, None)
        return newer

    def historize(self, previous):
        '''Records, in self.o's history, potential changes on historized fields.
           p_previous contains the values, before an update, of the historized
           fields, while p_self.o already contains the (potentially) modified
           values.'''
        o = self.o
        # Remove, from p_previous, any value that was not changed
        for name in list(previous.keys()):
            prev = previous[name]
            field = o.getField(name)
            curr = field.getValue(o, single=False)
            if (prev == curr) or ((prev is None) and (curr == '')) or \
               ((prev == '') and (curr is None)):
                del(previous[name])
                continue
            # The previous value may need to be formatted
            field.updateHistoryValue(o, previous)
        # Add the entry in the history (if not empty)
        if previous:
            self.add('Change', changes=previous)

    def getEvents(self, type, notBefore=None):
        '''Gets a subset of history events of some p_type. If specified, p_type
           must be the name of a concrete Event class or a list/tuple of such
           names.'''
        cond = 'not notBefore or (event.date >= notBefore)'
        return [event for event in self.iter(eventType=type, condition=cond, \
                                             context={'notBefore':notBefore})]

    def setEvents(self, other):
        '''Replace p_self's events with clones from p_other's events'''
        # Empty p_self's history first
        while len(self): del(self[0])
        # Create clones from p_other's events and insert them into p_self
        for event in other:
            self.append(event.clone())

    def getBatch(self, o, size):
        '''Returns the Batch instance allowing to navigate within p_o's
           history.'''
        # Compute the batch start and size
        size = int(size) if isinstance(size, str) else (size or 30)
        start = int(o.req.start or 0)
        return Batch(objects=self.data[start:start + size], total=len(self),
                     size=size, start=start, hook='history')

    def getAjaxData(self, batch):
        '''Gets data allowing to ajax-ask paginated history data'''
        params = {'start': batch.start, 'maxPerPage': batch.size}
        # Convert params into a JS dict
        params = sutils.getStringFrom(params)
        hook = batch.hook
        return "new AjaxData('%s/history/events','GET',%s,'%s')" % \
               (self.o.url, params, hook)

    def replaceLogin(self, old, new):
        '''Replace, in all events, login p_old with login p_new. Returns the
           number of replacements done.'''
        r = 0
        for event in self:
            if event.login == old:
                event.login = new
                r += 1
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path
import re, urllib.parse, subprocess

import appy.ui
from appy.tr import po
from appy.px import Px
from appy import Config
from appy.server import Server
from appy.model.base import Base
from appy.model.user import User
from appy.data import nativeNames
from appy.utils.dates import Date
from appy.model.group import Group
from appy.model.fields import Show
from appy.model.query import Query
from appy.database import Database
from appy.ui.layout import Layouts
from appy.utils.mail import sendMail
from appy.ui.template import Template
from appy.model.fields.ref import Ref
from appy.server.backup import Backup
from appy.model.searches import Search
from appy.model.fields.rich import Rich
from appy.model.fields import Initiator
from appy.model.carousel import Carousel
from appy.model.fields.phase import Page
from appy.model.page import Page as OPage
from appy.database.catalog import Catalog
from appy.model.fields.string import String
from appy.model.fields.action import Action
from appy.model.translation import Translation
from appy.model.fields.calendar import Calendar
from appy.model.fields.computed import Computed
from appy.server.context import AuthenticationContext

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RS_FG_KO   = 'Server cannot be restarted if running in the foreground.'
RESTART_S  = "Restarting server (wait %d'')..."

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Tool(Base):
    '''Base class for the Appy tool, a persistent instance storing configuration
       elements and more, like users, group and translations.'''

    # Make some modules and classes available and/or traversable via the tool
    ui = appy.ui
    Date = Date
    OPage = OPage
    Search = Search
    AuthContext = AuthenticationContext

    # While most users have generally the read permission on the tool and
    # translations, they must not be allowed to consult tool/view and sub-pages.
    # Traversing tool/view is thus by default restricted to those having the
    # write permission on the tool.
    traverse = Base.traverse.copy()
    traverse.update({'ui':True, 'guard':True, 'Search':True,
                     'AuthContext': 'Authenticated', 'Database':True,
                     'Server':True, 'view': 'perm:write'})

    Initiator = Initiator

    # The tool is not indexed by default
    indexable = False

    @staticmethod
    def update(class_):
        '''Hide pages being not necessary'''
        fields = class_.fields
        for name in ('title', 'record'):
            fields[name].page.show = False

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Tool initialisation
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def checkServers(self, config):
        '''Log the names of servers (LDAP, mail...) this app wants to connect
           to. Servers declared as "testable" are contacted to check the app's
           connection to it.'''
        servers = []
        # Browse connected servers
        for cfg in (config.security.ldap, config.mail, config.security.sso):
            if not cfg: continue
            # Initialise and check this server config
            cfg.init(self)
            enabled = 'enabled' if cfg.enabled else 'disabled'
            # Test the connection to the server
            status = ''
            if cfg.enabled and cfg.testable:
                status = cfg.test(self)
                if status: status = ', ' + status
            servers.append('%s (%s%s)' % (cfg, enabled, status))
        if servers:
            self.log('Server(s) %s configured.' % ', '.join(servers))

    def init(self, handler, poFiles):
        '''Ensure all required tool sub-objects are present and coherent'''
        # Get the root database object from the current request p_handler
        root = handler.dbConnection.root
        # Create the default users if they do not exist
        for login, roles in User.defaultUsers.items():
            # Does this user exist in the database ?
            if login in root.objects:
                # Yes. Do not create it. But ensure, for "anon" and "system",
                # that they have no email.
                user = root.objects.get(login)
                if login != 'admin': user.email = None
            else:
                # No. Create it.
                user = self.create('users', secure=False, id=login, login=login,
                            password=login, roles=roles)
                self.log('User "%s" created.' % login)

        # Ensure group "admins" exists. This group is granted role "Manager",
        # the role with the highest level of prerogatives in Appy.
        if 'admins' not in root.objects:
            self.create('groups', secure=False, id='admins', login='admins',
                        title='Administrators', roles=['Manager'])
            self.log('Group "admins" created.')

        # Load Appy "po" files if we need to inject their values into
        # Translation objects. app's "po" files are already in p_poFiles.
        # Also load ext's "po" files if an ext is defined.
        config = handler.server.config
        ui = config.ui
        load = ui.loadTranslationsAtStartup
        extFiles = None
        if load:
            appyFiles = po.load(pot=False, languages=ui.languages)
            if config.ext:
                extPath = '%s/tr' % __import__(config.ext).__path__[0]
                extFiles = po.load(path=Path(extPath), pot=False,
                                   languages=ui.languages)

        # Ensure a Translation object exists for every supported language
        for language in ui.languages:
            if language not in root.objects:
                # Create a Translation file
                tr = self.create('translations', secure=False, id=language,
                            title='%s (%s)' % (language, nativeNames[language]),
                            sourceLanguage=ui.sourceLanguage)
                tr.updateFromFiles(appyFiles, poFiles, extFiles or {})
            else:
                tr = root.objects[language]
                if load: tr.updateFromFiles(appyFiles, poFiles, extFiles or {})

        # Check connected servers
        self.checkServers(config)

        # Import an Appy 0 site when relevant
        peers = config.peers
        if peers and peers.appy0:
            peers.appy0.pump(self)

        # Define the home page
        Tool.default = getattr(Tool, config.ui.home)

        # Call method "onInstall" on the tool when available. This hook allows
        # an app to execute code on server initialisation.
        if hasattr(self, 'onInstall'): self.onInstall()

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Page "main"
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # Methods protecting visibility of tool pages and fields
    def forToolWriters(self):
        '''Some elements are only accessible to tool writers (ie Managers)'''
        if self.allows('write'): return Show.ER_

    def pageForToolWriters(self):
        '''Some tool pages are only accessible to tool writers (ie Managers)'''
        if self.allows('write'): return 'view'

    ta = {'label': 'Tool'}
    # Storing the Appy version on the tool (in the form of a string "x.y.z")
    # allows to trigger automatic migrations at server startup.
    appyVersion = String(show='view', **ta)

    # This hidden calendar is used for searches in "calendar" mode
    def calendarPreCompute(self, first, grid):
        '''Computes pre-computed information for the tool calendar'''
        # If this calendar is used as mode for a search, get this search
        req = self.req
        ctx = self.H().context
        mode = None
        if 'uiSearch' in ctx:
            # The search and related mode are already in the PX context
            mode = ctx.mode
        elif req.has_key('search') and req.has_key('className'):
            # Get the search and its mode from request keys
            className = req['className']
            tool = self.o
            search = self.o.getSearch(className, req['search'], ui=True)
            mode = search.Calendar(tool.getAppyClass(className), className,
                                   search, ctx.popup, tool)
            mode.init(req, ctx.dir)
        # Trigger the search via the mode
        if mode: mode.search(first, grid)
        return O(mode=mode)

    def calendarAdditionalInfo(self, date, preComputed):
        '''Renders the content of a given cell in the calendar used for
           searches' calendar mode.'''
        info = preComputed.mode.dumpObjectsAt(date)
        if info: return info

    calendar = Calendar(preCompute=calendarPreCompute, show=False,
                        additionalInfo=calendarAdditionalInfo, **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Page "users"
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    userPage = Page('users', show=pageForToolWriters, label='Tool_page_users',
                    icon='pageUsers.svg')

    # The "main" page, reserved to technical elements, may be hidden by apps
    def getDefaultViewPage(self): return 'users'

    # Ref(User) will maybe be transformed into Ref(AppUser)
    users = Ref(User, multiplicity=(0,None), add=True, link=False,
      composite=True, back=Ref(attribute='toTool', show=False, layouts='f'),
      page=userPage, queryable=True, queryFields=('searchable','login','roles'),
      show=forToolWriters, showHeaders=True, actionsDisplay='right',
      shownInfo=User.listColumns, **ta)

    def getUserName(self, login=None, normalized=False):
        '''Gets the user name corresponding to p_login (or the currently logged
           user if None), or the p_login itself if the user does not exist
           anymore. If p_normalized is True, special chars in the first and last
           names are normalized.'''
        if not login:
            user = self.user
        else:
            user = self.search1('User', login=login)
            if not user: return login
        return user.getTitle(normalized=normalized)

    def doSynchronizeExternalUsers(self):
        '''Synchronizes the local User copies with a distant LDAP user base'''
        cfg, context = self.config.security.getLdap()
        if not cfg: raise Exception('LDAP config not found.')
        message = cfg.synchronizeUsers(self, sso=context)
        return True, message

    def showSynchronizeUsers(self):
        '''Show this button only if a LDAP connection exists and is enabled'''
        cfg, context = self.config.security.getLdap()
        if cfg and cfg.enabled: return 'buttons'

    synchronizeExternalUsers = Action(action=doSynchronizeExternalUsers,
      show=showSynchronizeUsers, confirm=True, page=userPage, **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Page "groups"
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # Ref(Group) will maybe be transformed into Ref(AppGroup)
    groups = Ref(Group, multiplicity=(0,None), add=True, link=False,
      composite=True, back=Ref(attribute='toTool', show=False,layouts='f'),
      page=Page('groups', show=pageForToolWriters, label='Tool_page_groups',
                icon='groups.svg'),
      show=forToolWriters, queryable=True, queryFields=('title', 'login'),
      showHeaders=True, actionsDisplay='inline',
      shownInfo=('title', 'login*20%|', 'roles*20%|'), **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Page "translations"
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    pt = Page('translations', show=pageForToolWriters,
              label='Tool_page_translations', icon='pageTranslations.svg')

    # Ref(Translation) will maybe be transformed into Ref(AppTranslation)
    translations = Ref(Translation, multiplicity=(0,None), add=False,
      link=False, composite=True, show='view', page=pt, actionsDisplay='inline',
      back=Ref(attribute='toTool', show=False, layouts='f'), **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Page "pages"
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def showPagePages(self):
        '''Show page "pages" if the logged user is allowed to write the tool'''
        # The page may be edited only if there is at least one page defined (in
        # order to select it as default page).
        if not self.allows('write'): return
        return 'view' if self.isEmpty('pages') else True

    pp = Page('pages', show=showPagePages, label='Tool_page_pages',
              icon='pagePages.svg')

    pages = Ref(OPage, multiplicity=(0,None), add=True, link=False,
                composite=True, show=Show.VX, actionsDisplay='inline', page=pp,
                back=Ref(attribute='toTool', show=False, layouts='f'),
                numbered=True, showHeaders=True, shownInfo=OPage.listColumns,
                **ta)

    # The page, among "pages", being defined as the default one
    defaultPage = Ref(OPage, add=False, link=True, render='links', page=pp,
                      back=Ref(attribute='toTool2', show=False, layouts='f'),
                      show=lambda o: not o.isEmpty('pages'),
                      select=lambda o: o.pages, **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Page "carousels"
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    pc = Page('carousels', show=lambda o: 'view' if o.allows('write') else None,
              label='Tool_page_carousels')

    carousels = Ref(Carousel, multiplicity=(0,None), add=True, link=False,
                    composite=True, show=Show.VX, actionsDisplay='inline',
                    back=Ref(attribute='toTool', show=False, layouts=Layouts.f),
                    layouts=Layouts.dv, page=pc, showHeaders=True, **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Page "queries"
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    pq = Page('queries', show=lambda o: 'view' if o.allows('write') else None,
              label='Tool_page_queries')

    queries = Ref(Query, multiplicity=(0,None), add=True, link=False,
                  composite=True, show=Show.VX, actionsDisplay='inline',
                  back=Ref(attribute='toTool', show=False, layouts=Layouts.f),
                  layouts=Layouts.dv, page=pq, showHeaders=True,
                  shownInfo=Query.listColumns, **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  "admin" zone
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def forAdmin(self):
        '''Returns page visibility for pages from the admin zone'''
        return 'view' if self.user.hasRole('Manager') else None

    onlineUsers = Computed(method=lambda o: User.getOnline(o), layouts='f',
      show='view', page=Page('online', phase='admin', show=forAdmin,
                             label='Tool_page_online'), **ta)

    serverInfo = Computed(method=Server.view, layouts='f',
      page=Page('server', phase='admin', show=forAdmin,
                label='Tool_page_server'), **ta)

    databaseInfo = Computed(method=Database.view, layouts='f',
      page=Page('database', phase='admin', show=forAdmin,
                label='Tool_page_database'), **ta)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Main methods
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def sendMail(self, to, subject, body, attachments=None, replyTo=None):
        '''Sends a mail. See doc for appy.utils.mail.sendMail'''
        sendMail(self.config.mail, to, subject, body, attachments=attachments,
                 log=self.log, replyTo=replyTo)

    def computeHomePage(self):
        '''Compute the page that is shown to the user when he hits the app'''
        # Get the currently logged user
        user = self.user
        isAnon = user.isAnon()
        if isAnon and self.config.ui.discreetLogin:
            return '%s/public' % self.url
        # If the app defines a tool method names "getHomePage", call it
        r = None
        if hasattr(self, 'getHomePage'):
            r = self.getHomePage()
        if not r:
            # Bring anonymous users to tool/home, Managers to tool/view and
            # others to user/home.
            if isAnon:
                r = '%s/home' % self.url
            elif user.hasRole('Manager'):
                r = '%s/view' % self.url
            else:
                r = user.url
        return r

    def computeHomeObject(self, user, popup=False):
        '''The concept of "home object" is the object where the user must "be",
           even if he is "nowhere". For example, if the user is on a search
           screen, there is no contextual object (the tool excepted). In this
           case, if we have a home object for him, we will use it as contextual
           object, and its portlet menu will nevertheless appear: the user will
           not have the feeling of being lost.'''
        # If we are in the popup, we do not want any home object in the way
        if popup: return
        if hasattr(self, 'getHomeObject'):
            # If the app defines a method "getHomeObject", call it
            r = self.getHomeObject()
        else:
            # For managers, the home object is the tool. For others, there is
            # no default home object.
            if user.hasRole('Manager'): return self

    def getGoogleAnalyticsCode(self, handler, config):
        '''If the config defines a Google Analytics ID, this method returns the
           Javascript code to be included in every page, allowing Google
           Analytics to work.'''
        # Disable Google Analytics when we are in debug mode
        if handler.server.mode == 'fg': return
        # Disable Google Analytics if no ID is found in the config.
        gaId = config.googleAnalyticsId
        if not gaId: return
        # Google Analytics must be enabled: return the chunk of Javascript
        # code specified by Google.
        code = "var _gaq = _gaq || [];\n" \
               "_gaq.push(['_setAccount', '%s']);\n" \
               "_gaq.push(['_trackPageview']);\n" \
               "(function() {\n" \
               "  var ga = document.createElement('script'); " \
               "ga.type = 'text/javascript'; ga.async = true;\n" \
               "  ga.src = ('https:' == document.location.protocol ? " \
               "'https://ssl' : 'http://www') + " \
               "'.google-analytics.com/ga.js';\n" \
               "  var s = document.getElementsByTagName('script')[0]; " \
               "s.parentNode.insertBefore(ga, s);\n" \
               "})();\n" % gaId
        return code

    def formatDate(self, date, format=None, withHour=True, language=None):
        '''Check doc in called method'''
        if not date: return
        return self.config.ui.formatDate(self, date, format, withHour, language)

    def restart(self, wait=5):
        '''Restarts the currently running Appy server (S), by forking another
           process that will wait p_wait seconds before restarting S.'''
        # This cannot be done for servers running in the foreground
        if self.H().server.mode == 'fg':
            self.log(RS_FG_KO, type='warning')
            return
        # Fork a process that will launch a "./site restart" command. Waiting
        # p_wait seconds may give enough time to the Appy server for committing
        # ongoing transactions.
        sitePath = self.config.server.sitePath
        args = [str(sitePath / 'bin' / 'site'), 'restart']
        if wait:
            args.append('-w')
            args.append(str(wait))
        self.log(RESTART_S % wait)
        subprocess.Popen(args, stdout=None, stdin=None, stderr=None)

    def backup(self):
        '''Method allowing to trigger the backup of a Appy site'''
        return Backup(self).run()

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  PXs
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # PX "home" is the default page shown for a standard Appy site, excepted
    # when discreet logins are enabled. In this latter case, PX "public",
    # hereafter, is more appropriate.
    home = Px('''
     <!-- Redirect the user when appropriate -->
     <x if="(not isAnon or cfg.discreetLogin) and not req.stay"
        var2="x=tool.goto(tool.computeHomePage())"></x>

     <!-- Links -->
     <div class=":'%sTop' % _px_.name">:Template.pxTemplateLinks</div>

     <!-- The home logo and text -->
     <div if="isAnon and _px_.name == 'home'" class="homeText">
       <img src=":url('homeLogo')"/>
       <!-- The home text must not be shown here, if shown on a public page -->
       <x if="not tool.isEmpty('defaultPage') and \
              not cfg.discreetLogin">::tool.defaultPage.content</x>
     </div>

     <!-- The backgrounds for px "homes" -->
     <div if="_px_.name == 'homes'" class="homes">
      <div for="fileName, info in cfg.homesBackgrounds.items()"
           var2="base,width=info; bg=tool.buildUrl(fileName, base=base,bg=True)"
           style=":'%s;width:%s' % (bg, width)"></div>
     </div>''',

     css='''
      .homeText { position:fixed; left:45px; top:25px; color:|homeTextColor| }
      .homeText h1 { font-size: 300%; margin-left: -5px }
      .homeText h2 { font-size: 200% }
      .homeText p { font-size: 120%; margin: 0 }
      .homeText a, .homeText a:visited { color: #ffd8eb }
      .homeTop { position:fixed; top:20px; right:20px; color:|headerColor| }
     ''',

     template=Template.px, hook='content', name='home')

    # A variant of page "home", allowing to define several backgrounds, rendered
    # one besides each other.
    homes = Px(home.content, template=Template.px, hook='content', name='homes',
     css='''.homes { display:flex; width:100%; height:100% }
            .homes div { background-size: cover }
            .homesTop { position:absolute; text-align:center; left:0; right:0;
                        top:15px; z-index:10 }''')

    # PX "public" shows one of the root pages (rp), or the default one if none
    # of them is specified. When discreet logins are enabled, the public page
    # is shown, loaded with the default page.
    public = Px('''
     <!-- The home text -->
     <div class="public"
          var="layout='view';
               root=tool.getObject(req.rp) if req.rp else tool.defaultPage">
      <x if="root">
       <!-- Display links on top of the 1st page, if the header is not shown -->
       <div if="not showHeader"
            class="publicTop">:Template.pxTemplateLinks</div>

       <!-- Display the root page and all pages chained to it -->
       <x for="o in root.getChain()"
          var2="x=o.initialiseLocalPage(req)">:o.view</x>
      </x>
      <x if="not root">::_('no_default_page')</x>
     </div>''',

     css='''
      .publicTop { position:fixed; top:20px; right:20px; color:|headerColor| }
      .publicTop .img { width:27px }
      .publicTop .topIcons { display:inline }
     ''',

     template=Template.px, hook='content', name='public')

    # Disable tool removal
    def mayDelete(self): return
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re
from DateTime import DateTime
from appy.utils import string as sutils

# Importing database operators is required when evaluating criteria expressions
from appy.database.operators import and_, or_, in_, not_

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Criteria:
    '''Represents a set of search criteria manipulated from the UI'''

    def __init__(self, tool):
        self.tool = tool
        # This attribute will store the dict of search criteria, ready to be
        # injected in a Search class for performing a search in the catalog.
        self.criteria = None

    @classmethod
    def readFromRequest(class_, handler):
        '''Unmarshalls, from request key "criteria", a dict that was marshalled
           from a dict similar to the one stored in attribute "criteria" in
           Criteria instances.'''
        # Get the cached criteria on the handler if found
        cached = handler.cache.criteria
        if cached: return cached
        # Criteria may be absent from the request
        criteria = handler.req.criteria
        if not criteria: return
        # Criteria are present but not cached. Get them from the request,
        # unmarshal and cache them.
        r = eval(criteria)
        handler.cache.criteria = r
        return r

    @classmethod
    def highlight(class_, handler, text):
        '''Highlights parts of p_text if we are in the context of a search whose
           keywords must be highlighted.'''
        # Must we highlight something ?
        criteria = class_.readFromRequest(handler)
        if not criteria or ('searchable' not in criteria): return text
        # Highlighting operators is not supported yet
        keywords = criteria['searchable']
        if not isinstance(keywords, str): return text
        # Highlight every variant of every keyword. Use preferably the words as
        # encoded by the user instead of their normalized version.
        keywords = handler.req.w_searchable or keywords
        for word in keywords.strip().split():
            for variant in (word, word.capitalize(), word.lower()):
                highlighted = '<span class="highlight">%s</span>' % variant
                text = re.sub('(?<= |\(|\>)?%s' % variant, highlighted, text)
        return text

    def getFromRequest(self, class_):
        '''Retrieve search criteria from the request after the user has filled
           an advanced search form and store them in p_self.criteria.'''
        r = {}
        req = self.tool.req
        # Retrieve criteria from the request
        for name, value in req.items():
            # On search form, every search field is prefixed with "w_"
            if not name.startswith('w_'): continue
            name = name[2:]
            # Get the corresponding field
            field = class_.fields.get(name)
            # Ignore this value if it is empty or if the field is inappropriate
            # for a search.
            if not field or not field.indexed or field.searchValueIsEmpty(req):
                continue
            # We have a(n interval of) value(s) that is not empty for a given
            # field. Get it.
            r[name] = field.getSearchValue(req)
        # Complete criteria with Ref info if the search is restricted to
        # referenced objects of a Ref field.
        info = req.ref
        if info: r['_ref'] = info
        self.criteria = r

    def asString(self):
        '''Returns p_self.criteria, marshalled in a string'''
        return sutils.getStringFrom(self.criteria, stringify=False, c='"')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path
from DateTime import DateTime
import io, os, os.path, time, shutil, base64

from appy.px import Px
from appy import utils
from appy.ui.layout import Layouts
from appy.model.fields import Field
from appy.model.utils import Object
from appy.server.static import Static
from appy.utils import path as putils
from appy.utils import string as sutils
from appy.pod.converter import Converter
from appy.xml.unmarshaller import UnmarshalledFile

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
BAD_FTUPLE = 'This is not the way to set a file. You can specify a 2-tuple ' \
             '(fileName, fileContent) or a 3-tuple (fileName, fileContent, ' \
             'mimeType).'
CONV_ERR   = 'An error occurred. %s'
PATH_KO    = 'Missing absolute disk path for %s.'
RESIZED    = '%s resized to %spx.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class FileInfo:
    '''A FileInfo instance holds metadata about a file on the filesystem.

       For every File field, we will store a FileInfo instance in the dabatase;
       the real file will be stored in the Appy/ZODB database-managed
       filesystem.

       This is the primary usage of FileInfo instances. FileInfo instances can
       also be used every time we need to manipulate a file. For example, when
       getting the content of a Pod field, a temporary file may be generated and
       you will get a FileInfo that represents it.
    '''
    BYTES = 50000
    NOT_FOUND = 'File "%s" was not found.'

    def __init__(self, fsPath, inDb=True, uploadName=None):
        '''p_fsPath is the path of the file on disk.
           - If p_inDb is True, this FileInfo will be stored in the database and
             will hold metadata about a File field whose content will lie in the
             database-controlled filesystem. In this case, p_fsPath is the path
             of the file *relative* to the root DB folder. We avoid storing
             absolute paths in order to ease the transfer of databases from one
             place to the other. Moreover, p_fsPath does not include the
             filename, that will be computed later, from the field name.

           - If p_inDb is False, this FileInfo is a simple temporary object
             representing any file on the filesystem (not necessarily in the
             db-controlled filesystem). For instance, it could represent a temp
             file generated from a Pod field in the OS temp folder. In this
             case, p_fsPath is the absolute path to the file, including the
             filename. If you manipulate such a FileInfo instance, please avoid
             using methods that are used by Appy to manipulate
             database-controlled files (like methods getFilePath, removeFile,
             writeFile or copyFile).'''
        self.fsPath = fsPath
        self.fsName = None # The name of the file in fsPath
        self.uploadName = uploadName # The name of the uploaded file
        self.size = 0 # Its size, in bytes
        self.mimeType = None # Its MIME type
        self.modified = None # The last modification date for this file
        # Complete metadata if p_inDb is False. p_inDb is not stored as is:
        # checking if self.fsName is the empty string is equivalent.
        if not inDb:
            self.fsName = '' # Already included in self.fsPath
            self.setAttributes(self.fsPath)

    def setAttributes(self, path):
        '''Compute file attributes (size, MIME type and last modification date)
           for a file whose absolute p_path is passed and store these attributes
           on p_self.'''
        info = os.stat(path)
        self.size = info.st_size
        self.mimeType = putils.guessMimeType(path)
        self.modified = DateTime(info.st_mtime)

    def inDb(self):
        '''Does this FileInfo represent a file from the DB-controlled
           filesystem ?'''
        return bool(self.fsName)

    def getFilePath(self, o):
        '''Returns the Path instance corresponding to the file on disk that
           corresponds to this FileInfo instance.'''
        folder = o.getFolder()
        path = folder / self.fsName
        if not path.is_file():
            # It may already have been deleted by a failed transaction. Try to
            # get a copy we may have made in the OS temp folder.
            path = Path(putils.getOsTempFolder()) / folder.name / self.fsName
        return path

    def removeFile(self, folder, removeEmptyFolders=False):
        '''Removes the file from the filesystem'''
        path = folder / self.fsName
        if path.is_file():
            # If the current ZODB transaction is re-triggered, the file may
            # already have been deleted.
            path.unlink()
        # Don't leave empty folders on disk. So delete folder and parent folders
        # if this removal leaves them empty (unless p_removeEmptyFolders is
        # False).
        if removeEmptyFolders:
            putils.FolderDeleter.deleteEmpty(str(folder))

    def normalizeFileName(self, name):
        '''Normalizes file p_name.'''
        return name[max(name.rfind('/'), name.rfind('\\'), name.rfind(':'))+1:]

    def getShownSize(self): return putils.getShownSize(self.size)

    def replicateFile(self, src, dest):
        '''p_src and p_dest are open file handlers. This method copies content
           of p_src to p_dest and returns the file size. Note that p_src can
           also be binary data as bytes.'''
        size = 0
        src = io.BytesIO(src) if isinstance(src, bytes) else src
        while True:
            chunk = src.read(self.BYTES)
            if not chunk: break
            size += len(chunk)
            dest.write(chunk)
        return size

    def getBase64(self, o=None):
        '''Returns the file content, as a base64-encoded string'''
        path = self.getFilePath(o) if o else self.fsPath
        f = open(path, 'rb')
        r = base64.b64encode(f.read())
        f.close()
        return r

    def getExtension(self):
        '''Get the file extension from the MIME type or from the upload name'''
        if self.mimeType in mimeTypesExts:
            return mimeTypesExts[self.mimeType]
        elif self.uploadName:
            parts = os.path.splitext(self.uploadName)
            if len(parts) > 1:
                return parts[-1][1:]

    def writeResponse(self, handler, path=None, disposition='attachment',
                      cache=False):
        '''Returns the content the file in the HTTP response'''
        if not self.fsName:
            # Not in-database file: we have its full path
            path = self.fsPath
        elif not path:
            # An in-database file for which we do not have the full path
            raise Exception(PATH_KO % self.fsName)
        # Produce the file via the Static class
        Static.write(handler, path, None, fileInfo=self,
                     disposition=disposition, enableCache=cache)

    def writeFile(self, name, value, folder, config=None):
        '''Writes a file to the filesystem, from p_value that can be:
           - an Object instance (coming from a HTTP post);
           - another ("not-in-DB") FileInfo instance;
           - a tuple (fileName, content, mimeType): see method File.store.'''
        # Determine the file's MIME type
        isO = isinstance(value, Object) or isinstance(value, UnmarshalledFile)
        if isO:
            mimeType = value.type
            niceName = value.name
        elif isinstance(value, FileInfo):
            mimeType = value.mimeType
            niceName = value.uploadName
        else:
            mimeType = value[2]
            niceName = value[0]
        self.mimeType = mimeType or putils.defaultMimeType
        # Force the file to be written on disk with the field name, serving as
        # an identifier within the database p_folder.
        ext = utils.mimeTypesExts.get(self.mimeType) or 'bin'
        self.fsName = '%s.%s' % (name, ext)
        self.uploadName = self.normalizeFileName(niceName)
        # Write the file on disk (and compute/get its size in bytes)
        path = str(folder / self.fsName)
        f = open(path, 'wb')
        if isO:
            content = value.value
            self.size = len(content)
            f.write(content)
        elif isinstance(value, FileInfo):
            fsPath = value.fsPath
            if not fsPath.startswith('/'):
                # The p_value is a "in-db" FileInfo instance. If p_config is
                # there, we can reconstitute its absolute path.
                fsPath = config.database.binariesFolder / fsPath / value.fsName
                fsPath = str(fsPath)
            src = open(fsPath, 'rb')
            self.size = self.replicateFile(src, f)
            src.close()
        else:
            # Write value[1] on disk
            content = value[1]
            if isinstance(content, bytes):
                self.size = len(content)
                f.write(content)
            else:
                # An open file handler
                self.size = self.replicateFile(content, f)
        f.close()
        self.modified = DateTime()

    def copyFile(self, fieldName, filePath, dbFolder):
        '''Copies the "external" file stored at p_filePath in the db-controlled
           file system, for storing a value for p_fieldName.'''
        # Set names for the file
        filePath = str(filePath) if isinstance(filePath, Path) else filePath
        name = self.normalizeFileName(filePath)
        self.uploadName = name
        self.fsName = '%s%s' % (fieldName, os.path.splitext(name)[1])
        # Set mimeType
        self.mimeType = putils.guessMimeType(filePath)
        # Copy the file
        fsName = dbFolder / self.fsName
        shutil.copyfile(filePath, fsName)
        self.modified = DateTime()
        self.size = os.stat(fsName).st_size

    def dump(self, o, filePath=None, format=None):
        '''Exports this file to disk (outside the db-controlled filesystem).
           The tied p_o(bject) is required. If p_filePath is specified, it
           is the path name where the file will be dumped; folders mentioned in
           it must exist. If not, the file will be dumped in the OS temp folder.
           The absolute path name of the dumped file is returned. If an error
           occurs, the method returns None. If p_format is specified,
           LibreOffice will be called for converting the dumped file to the
           desired format.'''
        if not filePath:
            filePath = '%s/file%f.%s' % (putils.getOsTempFolder(), time.time(),
                                         self.fsName)
        # Copies the file to disk
        shutil.copyfile(self.getFilePath(o), filePath)
        if format:
            # Convert the dumped file using LibreOffice
            out, err = Converter(filePath, format).run()
            # Even if we have an "error" message, it could be a simple warning.
            # So we will continue here and, as a subsequent check for knowing if
            # an error occurred or not, we will test the existence of the
            # converted file (see below).
            os.remove(filePath)
            # Get the name of the converted file
            baseName, ext = os.path.splitext(filePath)
            if ext == ('.%s' % format):
                filePath = '%s.res.%s' % (baseName, format)
            else:
                filePath = '%s.%s' % (baseName, format)
            if not os.path.exists(filePath):
                o.log(CONV_ERR % err, type='error')
                return
        return filePath

    def resize(self, folder, width, o):
        '''Resize this image if it is a resizable image'''
        if self.mimeType not in File.resizableImages: return
        # Get the absolute path to the file on disk
        path = folder / self.fsName
        # Get the width, in pixels, that will be used to resize it
        width = sutils.Normalize.digit(str(width))
        utils.convert(path, '-resize %sx%s>' % (width, width))
        o.log(RESIZED % (path, width))
        # (re)compute p_self's attributes, after resizing
        self.setAttributes(str(path))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class File(Field):
    '''Field allowing to upload a binary file'''

    # Some elements will be traversable
    traverse = Field.traverse.copy()

    # MIME types for images
    imageMimeTypes = ('image/png', 'image/jpeg', 'image/gif')
    # MIME types for resizable images
    resizableImages = imageMimeTypes

    class Layouts(Layouts):
        '''File-specific layouts'''
        b = Layouts(edit='lrv-f', view='l-f')
        bg = Layouts(edit='frvl', view='fl')

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this File p_field'''
            return class_.bg if field.inGrid() else class_.b

    view = cell = Px('''
      <x>::field.getDownloadLink(name, o, layout)</x>''')

    edit = Px('''
     <x var="fName='%s_file' % name">
      <x if="value">
       <x>:field.view</x><br/>
       <!-- Keep the file unchanged -->
       <input type="radio" value="keep" checked=":bool(value)"
              name=":'%s_action' % name" id=":'%s_keep' % name"
              onclick=":'document.getElementById(%s).disabled=true'% q(fName)"/>
       <label lfor=":'%s_keep' % name">:_('keep_file')</label><br/>
       <!-- Delete the file -->
       <x if="not field.required">
        <input type="radio" value="delete"
               name=":'%s_action' % name" id=":'%s_delete' % name"
               onclick=":'document.getElementById(%s).disabled=true'%q(fName)"/>
        <label lfor=":'%s_delete' % name">:_('delete_file')</label><br/>
       </x>
       <!-- Replace with a new file -->
       <input type="radio" value="replace"
              checked=":not value and 'checked' or None"
              name=":'%s_action' % name" id=":'%s_upload' % name"
              onclick=":'document.getElementById(%s).disabled=false'%q(fName)"/>
       <label lfor=":'%s_upload' % name">:_('replace_file')</label><br/>
      </x>
      <!-- The upload field -->
      <input type="file" name=":fName" id=":fName" style=":field.getStyle()"
             onChange=":field.getJsOnChange()"/>
      <script var="isDisabled=not value \
             and 'false' or 'true'">:'document.getElementById(%s).disabled=%s'%\
                                     (q(fName), isDisabled)</script></x>''')

    search = ''

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, readPermission='read', writePermission='write', width=None,
      height=None, inputWidth=None, maxChars=None, colspan=1, master=None,
      masterValue=None, focus=False, historized=False, mapping=None,
      generateLabel=None, label=None, isImage=False, downloadAction=None,
      sdefault='', scolspan=1, swidth=None, sheight=None, view=None, cell=None,
      edit=None, xml=None, translations=None, render=None, icon='paperclip',
      disposition='attachment', nameStorer=None, cache=False, resize=False,
      thumbnail=None, viewWidth=None):
        # This boolean is True if the file is an image
        self.isImage = isImage
        # "downloadAction" can be a method called every time the file is
        # downloaded. The method gets, as single arg, the FileInfo instance
        # representing the downloaded file.
        self.downloadAction = downloadAction
        # If "render" is "icon", the file will be rendered as an icon
        # (self.icon) on "buttons" and "result" layouts.
        self.render = render
        # Icon to use when this file is rendered as an icon
        self.icon = icon
        # In "nameStorer", you can specify the name of another field that will
        # store the file name. This field must be a String belonging to the same
        # class as p_self. As soon as, in the UI, a file is selected in p_self's
        # widget, its name will be copied into the nameStorer field, without the
        # extension, only if this latter is not empty.
        self.nameStorer = nameStorer
        # When downloading the file, the browser may try to open it within the
        # browser ("inline" disposition) or save it to the disk ("attachment"
        # disposition). This latter is the default. You may also specify a
        # method returning one of these 2 valid values.
        self.disposition = disposition
        # By default, caching is disabled for this field: the browser will not
        # be able to cache the file stored in it. This is the default, for
        # security reasons. If you decide the browser can cache the file stored
        # here, set the following attribute to True. Attribute p_cache can also
        # store a method that must return a boolean value.
        self.cache = cache
        # Attribute "width" is used to specify the width of the image or
        # document. The width of the input field allowing to upload the file can
        # be defined in attribute "inputWidth".
        self.inputWidth = inputWidth
        # If attribute "resize" is False (the default), the image will be
        # previewed in the UI with dimensions as defined in attributes "width"
        # and "height", but without being effectively resized. If "resize" is
        # True, when uploading the file, it will be resized to self.width,
        # keeping the width / height ratio.
        self.resize = resize
        # If you place here the name of another File field, it will be used to
        # store a thumbnail of the image stored in the current field. The
        # thumbnail's dimension is defined, as usual, via attribute "width" of
        # the File field storing it. Everytime a document is uploaded in p_self,
        # if a thumbnail is defined in its "thumbnail" attribute, the thumbnail
        # field will be updated and will store the same image as the one stored
        # in p_self, resized accordingly (provided you have defined it as
        # "resizable").
        self.thumbnail = thumbnail
        # When the uploaded file is an image, on "view" and "cell" layouts, you
        # may want to display it smaller than its actual size. In order to do
        # that, specify a width (as a string, ie '700px') in the following
        # attribute. The attribute may also hold a method returning the value.
        self.viewWidth = viewWidth
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, False, True, None, None, False,
          None, readPermission, writePermission, width, height, None, colspan,
          master, masterValue, focus, historized, mapping, generateLabel, label,
          sdefault, scolspan, swidth, sheight, True, False, view, cell, edit,
          xml, translations)

    def getRequestValue(self, o, requestName=None):
        name = requestName or self.name
        return o.req['%s_file' % name]

    def getRequestSuffix(self): return '_file'

    def getCopyValue(self, o):
        '''Create a copy of the FileInfo instance stored for p_o for this field.
           This copy will contain the absolute path to the file on the
           filesystem. This way, the file may be read independently from p_o
           (and copied somewhere else).'''
        info = self.getValue(o)
        if not info: return
        # Create a "not-in-DB", temporary FileInfo
        return FileInfo(info.getFilePath(o), inDb=False,
                        uploadName=info.uploadName)

    def isEmptyValue(self, o, value):
        '''Must p_value be considered as empty ?'''
        if value: return
        # If "keep", the value must not be considered as empty
        return o.req['%s_action' % self.name] != 'keep'

    def getJsOnChange(self):
        '''Gets the JS code for updaing the name storer when defined'''
        storer = self.nameStorer
        if not storer: return ''
        return 'updateFileNameStorer(this, "%s")' % storer

    def getStyle(self):
        '''Get the content of the "style" attribute of the "input" tag on the
           "edit" layout for this field.'''
        return 'width: %s' % self.inputWidth if self.inputWidth else ''

    def isRenderable(self, layout):
        '''A file with 'icon' rendering is potentially renderable everywhere'''
        if self.render == 'icon': return True
        return layout != 'buttons'

    def getDownloadLink(self, name, o, layout):
        '''Gets the HTML code for downloading the file as stored in field
           named p_name on p_o.'''
        value = self.getValueIf(o, name, layout)
        # Display an empty value
        if not value: return '' if layout == 'cell' else '-'
        # On "edit", simply repeat the file title
        if layout == 'edit': return value.uploadName
        # Get the "file title", derived from the upload name and size
        size = value.getShownSize()
        title = "%s - %s" % (value.uploadName, size)
        # Build the URL for downloading or displaying the file
        url = '%s/%s/download' % (o.url, name)
        # For images, display them directly
        if self.isImage:
            # Define a max width when relevant
            viewWidth = self.getAttribute(o, 'viewWidth')
            css = ' style="max-width: %s"' % viewWidth if viewWidth else ''
            return '<img src="%s" title="%s"%s/>' % (url, title, css)
        # For non-images, display a link for downloading it, as an icon when
        # relevant.
        if self.render == 'icon':
            content = '<img src="%s" title="%s"/>' % \
                      (o.buildUrl(self.icon), title)
            # On "view", we have place, so display "title" besides the icon
            suffix = title if layout == 'view' else ''
        else:
            # Display textual information only
            content = title
            suffix = ''
        # Style the suffix
        if suffix: suffix = '<span class="refLink">%s</span>' % suffix
        return '<a href="%s">%s%s</a>' % (url, content, suffix)

    def isCompleteValue(self, o, value):
        '''Always consider the value being complete, even if empty, in order to
           be able to check if, when the user has checked the box indicating
           that he will replace the file with another one, he has uploaded
           another file.'''
        return True

    def validateValue(self, o, value):
        '''Ensure p_value is valid'''
        if not value and (o.req['%s_action' % self.name] == 'replace'):
            # The user has selected "replace the file with a new one" but has
            # not uploaded a new file.
            return o.translate('file_required')
        # Check that, if self.isImage, the uploaded file is really an image
        elif value and self.isImage:
            if value.type not in File.imageMimeTypes:
                return o.translate('image_required')

    def store(self, o, value):
        '''Stores the p_value that represents some file'''
        # p_value can be:
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  a | an instance of class appy.model.utils.Object, built from data
        #    | originating from a HTTP POST request, having attributes:
        #    |  - "name",    containing the name of an uploaded file;
        #    |  - "type",    containing the MIME type of this file;
        #    |  - "value",   containing its binary content;
        #    | or an instance of appy.xml.unmarshalled.UnmarshalledFile, having
        #    | the same attributes, representing a file unmarshalled from XML
        #    | content.
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  b | a string or pathlib.Path instance. In this case, it represents
        #    | the path of a file on disk;
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  c | a 2-tuple (name, content) where:
        #    | - "name"     is the name of the file (ie "myFile.odt")
        #    | - "content"  is the binary or textual content of the file or an
        #    |              open file handler.
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  d | a 3-tuple (name, content, mimeType) where
        #    | - "name" and "content" have the same meaning than above;
        #    | - "mimeType" is the MIME type of the file.
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  e | a FileInfo instance, be it "in-DB" or not.
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        if value:
            # There is a new value to store. Get the folder on disk where to
            # store the new file.
            folder, relative = o.getFolder(create=True, withRelative=True)
            # Remove the previous file if it existed
            info = o.values.get(self.name)
            if info: info.removeFile(folder)
            # Store the new file. As a preamble, create a FileInfo instance.
            info = FileInfo(relative)
            if isinstance(value, Object) or isinstance(value, FileInfo) or \
               isinstance(value, UnmarshalledFile):
                # Cases (a) and (e)
                info.writeFile(self.name, value, folder, config=o.config)
            elif isinstance(value, str) or isinstance(value, Path):
                # Case (b)
                info.copyFile(self.name, value, folder)
            else:
                # Cases (c) and (d): extract file name, content and MIME type
                name = type = None
                if len(value) == 2:
                    name, content = value
                elif len(value) == 3:
                    name, content, type = value
                if not name:
                    raise Exception(BAD_FTUPLE)
                type = type or putils.guessMimeType(name)
                info.writeFile(self.name, (name, content, type), folder)
            # Store the FileInfo instance in the database
            o.values[self.name] = info
            # Resize the image when relevant
            if self.resize and self.width:
                width = self.getAttribute(o, 'width')
                info.resize(folder, width, o)
            # Update the thumbnail if defined
            fileName = str(folder / info.fsName)
            if self.thumbnail: o.getField(self.thumbnail).store(o, fileName)
        else:
            # Store value "None", excepted if we find, in the request, the
            # desire to keep the file unchanged.
            if o.req['%s_action' % self.name] != 'keep':
                # Delete the file on disk if it existed
                info = o.values.get(self.name)
                if info:
                    info.removeFile(o.getFolder(), removeEmptyFolders=True)
                    # Delete the FileInfo in the DB
                    del(o.values[self.name])

    traverse['download'] = 'perm:read'
    def download(self, o):
        '''Triggered when a file download is requested from the ui'''
        # Write the file in the HTTP response
        info = o.values.get(self.name)
        config = o.config.server.static
        handler = o.H()
        if info:
            # Return a 404 if the file does not exist
            path = info.getFilePath(o)
            if not path.is_file():
                Static.notFound(handler, config)
                return
            # Call the "download action" if specified
            if self.downloadAction:
                self.downloadAction(o, info)
                # In that case, a commit is required
                o.H().commit = True
            # Send the file to the browser
            info.writeResponse(handler, str(path),
                               disposition=self.getAttribute(o, 'disposition'),
                               cache=self.getAttribute(o, 'cache'))
        else:
            # Return a 404
            Static.notFound(handler, config)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Utility module related to URL manipulation'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import urllib.parse

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def split(url):
    '''Returns a tuple (url, params), where "url" is p_url whose parameters have
       been removed, and "params" is a dict containing these params, or None if
       there is no param in p_url.'''
    # Extract the various p_url parts
    parsed = urllib.parse.parse(url)
    # Return the URL as-is if it contains no parameter
    if not parsed.query: return url, None
    # Compute the dict of parameters
    r = {}
    for param in parsed.query.split('&'):
        if '=' in param:
            name, value = param.split('=', 1)
            r[name] = value
        else:
            r[name] = None
    # Re-build the URL parts into an URL, but without params
    url = '%s://%s%s' % (parsed.scheme, parsed.netloc, parsed.path)
    return url, r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def encode(params, ignoreNoneValues=True):
    '''Encode dict of p_params as a string of parameters ready to be
       incorporated into an URL. If p_ignoreNoneValues is True, every entry
       whose value is None is ignored.'''
    if not params: return params
    r = []
    for name, value in params.items():
        # Ignore entries whose value is None when appropriate
        if ignoreNoneValues and (value is None): continue
        # Convert the value to a str if not done yet
        value = value if isinstance(value, str) else str(value)
        # Add the encoded entry to the result
        r.append('%s=%s' % (name, value))
    return '&'.join(r)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
# Import stuff from appy.fields (and from a few other places too).
# This way, when an app gets "from appy.gen import *", everything is available.
import os.path
from appy.px import Px
from appy import Object
from appy.fields import Field
from appy.fields.pod import Pod

from appy.fields.text import Text
from appy.fields.info import Info
from appy.fields.date import Date
from appy.fields.file import File
from appy.fields.list import List
from appy.fields.dict import Dict
from appy.fields.rich import Rich
from appy.fields.page import Page
from appy.fields.hour import Hour
from appy.fields.workflow import *
from appy.fields.float import Float
from appy.fields.phase import Phase
from appy.fields.color import Color
from appy.fields.string import String
from appy.fields.action import Action
from appy.fields.custom import Custom
from appy.fields.switch import Switch
from appy.fields.boolean import Boolean
from appy.fields.integer import Integer
from appy.fields.ref import Ref, autoref
from appy.fields.computed import Computed
from appy.gen.layout import Table, Layouts
from appy.gen.monitoring import Monitoring
from appy.fields.group import Group, Column
from appy.fields.select import Select, Selection
from appy.fields.search import Search, UiSearch, Gridder
from appy.gen.utils import No, Tool, User, MessageException

# Make the following classes available here: people may need to override some
# of their PXs (defined as static attributes).
from appy.gen.wrappers import AbstractWrapper as BaseObject
from appy.gen.wrappers.ToolWrapper import ToolWrapper as BaseTool

# ------------------------------------------------------------------------------
class Ui:
    '''All configuration elements related to the user interface. Class Config
       below has a static attribute named "ui" holding this class.'''
    def __init__(self):
        # The top banner
        self.banner = 'banner.png'
        # If you place a function in the following attribute, it will be called
        # with, as args, the current PX and its context, and must return True
        # whenever the banner must be shown. If no function is set, the banner
        # will always be shown.
        self.bannerShow = None
        # Similar functions can be defined for determining visibility of other
        # UI elements.
        self.userStripShow = None
        self.footerShow = None
        # Fonts in use
        self.fonts = '"Lucida Grande","Lucida Sans Unicode",Arial,sans-serif'
        self.fontSize = '75%'
        # Among the fonts listed above, specify here, as a tuple, those being
        # Google fonts. That way, the corresponding "CSS include" will be
        # injected into all the pages from your app.
        self.googleFonts = ()
        # If you want to add specific CSS classes to some standard Appy parts,
        # specify a function in the following attribute. The function will
        # receive, as args, the name of the concerned part, the current PX and
        # its context; it must return the name of one or more CSS classes or
        # None when no class must be added. Currently defined parts are the
        # following.
        # ----------------------------------------------------------------------
        #   body      | The page "body" tag
        #   main      | The main zone of any page, behind the user strip
        # ----------------------------------------------------------------------
        self.css = None
        # Some input fields will get this background color once they will
        # contain erroneous content.
        self.wrongTextColor = '#f9edbe'
        # The tool may be configured in write-access only for a technical
        # reason, ie, for allowing user self-registration. Indeed, in that case,
        # anonymous users must be granted the ability to add a User instance in
        # Ref tool.users. In that case, we don't want to show the icon allowing
        # to access the tool to anyone having write-access to it. For these
        # cases, a specific function may be defined here for determining
        # showability of the tool's icon in the UI. This function will receive
        # the tool as unique arg.
        self.toolShow = None
        # Border style for tabs
        self.tabBorder = '1px solid #ff8040'

    def getBannerName(self, dir):
        '''If your site uses at least one RTL (right-to-left) language, you must
           propose a banner whose name is Ui.banner, suffixed with "rtl".'''
        res = self.banner
        if dir == 'rtl': res = '%srtl%s' % os.path.splitext(res)
        return res

    def _show(self, elem, px, context, inPopup):
        ''''As a general rule, always show UI p_elem, excepted in the popup or
            when a custom function prevents it.'''
        # Do not show it if we are in the popup
        if inPopup: return
        # A custom function may determine it. If no such function is defined,
        # show it.
        fun = getattr(self, '%sShow' % elem)
        return not fun and True or fun(px, context)

    def showBanner(self, px, context, inPopup):
        return self._show('banner', px, context, inPopup)

    def showUserStrip(self, px, context, inPopup):
        return self._show('userStrip', px, context, inPopup)

    def showFooter(self, px, context, inPopup):
        return self._show('footer', px, context, inPopup)

    def getClass(self, part, wide, px, context):
        '''Get the CSS classes that must be defined for some UI p_part, if
           any.'''
        # Start with the standard set of Appy classes that must set on p_part
        if part == 'main':
            r = wide and 'mainWide main rel' or 'main rel'
        else:
            r = ''
        # Add specific classes when relevant
        if self.css:
            add = self.css(part, px, context)
            if not add: return r
            if not r:
                r = add
            else:
                r = '%s %s' % (r, add)
        return r

    def getFontsInclude(self):
        '''If Google Fonts are in use, return the link to the CSS include
           allowing to use it.'''
        families = '|'.join(self.googleFonts)
        return 'https://fonts.googleapis.com/css?family=%s' % families

# ------------------------------------------------------------------------------
class Show:
    '''Provides some frequently used values for field's "show" attributes'''
    # As a convention, a trailing underscore indicates a negation

    # All layouts but edit. To use for users that can consult the field value
    # but cannot modify it. Also used for fields like some Ref fields that are
    # never manipulated via "edit" layouts.
    E_ = ('view', 'result', 'xml')
    # A variant, without "result"
    ER_ = ('view', 'xml')
    # A variant, with layout "buttons" instead of "result". "buttons"
    # corresponds, on lists, to the range of icons and/or buttons present in the
    # "title" column.
    E_B = ('view', 'buttons', 'xml')
    B = ('edit', 'view', 'buttons', 'xml')

    # All layouts but view. To use typically when, on view, the field value is
    # already shown as a part of some custom widget.
    V_ = ('edit', 'result', 'xml')
    # All layouts but view and edit. To use when you need both E_ and V_.
    VE_ = ('result', 'xml')
    # This set is used for showing workflow transitions in lists of objects
    TR = ('view', 'result')
    # This is, a.o., for custom widgets whose edit and view variants are not
    # that different, but that cannot be shown elsewhere (result, xml, etc).
    VE = ('view', 'edit')
    VX = ('view', 'xml')
    EX = ('edit', 'xml')

# ------------------------------------------------------------------------------
class Config:
    '''If you want to specify some configuration parameters for appy.gen and
       your application, please create a class named "Config" in the __init__.py
       file of your application and override some of the attributes defined
       here, ie:

       import appy.gen
       class Config(appy.gen.Config):
           langages = ('en', 'fr')
    '''
    # What skin to use for the web interface? Appy has 2 skins: the default
    # one (with a fixed width) and the "wide" skin (takes the whole page width).
    skin = None # None means: the default one. Could be "wide".
    # A reference to the Ui class (see above)
    ui = Ui()
    # For every language code that you specify in this list, appy.gen will
    # produce and maintain translation files.
    languages = ['en']
    # If languageSelector is True, on (almost) every page, a language selector
    # will allow to switch between languages defined in self.languages. Else,
    # the browser-defined language will be used for choosing the language
    # of returned pages.
    languageSelector = False
    # If "forceLanguage" is set, Appy will not take care of the browser
    # language, will always use the forced language and will hide the language
    # selector, even if "languageSelector" hereabove is True.
    forcedLanguage = None
    # When no translation is available in some language, Appy will fall back to
    # translations in this language.
    fallbackLanguage = 'en'
    # Show the link to the user profile in the user strip
    userLink = True
    # If you want to distinguish a test site from a production site, set the
    # "test" parameter to some text (lie "TEST SYSTEM" or
    # "VALIDATION ENVIRONMENT". This text will be shown on every page. This
    # parameter can also hold a function that will accept the tool as single
    # argument and returns the message.
    test = None

    @classmethod
    def getTestMessage(class_, tool):
        '''Returns the test message via attribute or method in "test"'''
        r = class_.test
        if callable(r): r = r(tool)
        return r

    # People having one of these roles will be able to create instances
    # of classes defined in your application.
    defaultCreators = ['Manager']
    # Roles in use in a Appy application are identified at generation time from
    # workflows or class attributes like "creators": it is not needed to declare
    # them somewhere. If you want roles that Appy will be unabled to detect, add
    # them in the following list. Every role can be a Role instance or a string.
    additionalRoles = []
    # What roles can unlock locked pages ? By default, only a Manager can do it.
    # You can place here roles expressed as strings or Role instances, global or
    # local.
    unlockers = ['Manager']
    # The "root" classes are those that will get their menu in the user
    # interface. Put their names in the list below. If you leave the list empty,
    # all gen-classes will be considered root classes (the default). If
    # rootClasses is None, no class will be considered as root.
    rootClasses = []
    # By default, instances of root classes will be stored in a database folder
    # named "data". You can specify an alternative base folder if the number of
    # created instances in this folder would become too high (maybe > 100 000).
    rootFolders = {}
    # Number of translations for every page on a Translation object
    translationsPerPage = 30
    # Language that will be used as a basis for translating to other
    # languages.
    sourceLanguage = 'en'
    # Activate or not the button on home page for asking a new password
    activateForgotPassword = True
    # Enable session timeout ?
    enableSessionTimeout = False
    # When a login is encoded by the user, a transform can be applied: upper,
    # lower or capitalize.
    loginTransform = None
    # If the following field is True, the login/password widget will be
    # discreet. This is for sites where authentication is not foreseen for
    # the majority of visitors (just for some administrators).
    discreetLogin = False
    # If the following attribute is True, a comment within an object's history
    # will be editable, either by a Manager or by the user having performed the
    # corresponding action.
    editHistoryComments = False
    # On user log on, you may want to ask them to choose their "authentication
    # context" in a list. Such a context may have any semantics, is coded as a
    # string and will be accessible on the Request object in attribute
    # "authContext". In order to activate authentication contexts, place here an
    # instance of a sub-class of appy.gen.authenticate::AuthenticationContext.
    # Please consult appy/gen/authenticate.py for more information.
    authContext = None
    # When using Ogone, place an instance of appy.gen.ogone.OgoneConfig in
    # the field below.
    ogone = None
    # When using Google analytics, specify here the Analytics ID
    googleAnalyticsId = None
    # Create a group for every global role ?
    groupsForGlobalRoles = False
    # When using a LDAP for authenticating users, place an instance of class
    # appy.shared.ldap.LdapConfig in the field below.
    ldap = None
    # When using, in front of an Appy server, a reverse proxy for authentication
    # for achieving single-sign-on (SSO), place an instance of
    # appy.shared.sso.SsoConfig in the field below.
    sso = None

    @classmethod
    def getLdap(class_):
        '''Returns the LDAP config if it exists'''
        # The LDAP config can be found either in global config attribute "ldap"
        # or in the context of a SSO system. This method returns a tuple
        #                     (ldapConfig, context)
        # "context" being None if the global LDAP config is returned, or being
        # the SSO config if the LDAP is the one embedded in this SSO config.
        if class_.ldap:
            r = class_.ldap, None
        else:
            sso = class_.sso
            if not sso or not sso.ldap:
                r = None, None
            else:
                r = sso.ldap, sso
        return r

    # When using a SMTP mail server for sending emails from your app, place an
    # instance of class appy.gen.mail.MailConfig in the field below.
    mail = None
    # For an app, the default folder where to look for static content for the
    # user interface (CSS, Javascript and image files) is folder "ui" within
    # this app.
    uiFolders = ['ui']
    # CK editor configuration. Appy integrates CK editor via CDN (see
    # http://cdn.ckeditor.com). Do not change "ckVersion" hereafter, excepted
    # if you are sure that the customized configuration files config.js,
    # contents.css and styles.js stored in appy/gen/ui/ckeditor will be
    # compatible with the version you want to use.
    ckVersion = '4.14.0'
    # ckDistribution can be "basic", "standard", "standard-all", "full" or
    # "full-all" (see doc in http://cdn.ckeditor.com).
    ckDistribution = 'standard'
    # CK toolbars are not configurable yet. So toolbar "Appy", defined in
    # appy/gen/ui/ckeditor/config.js, will always be used.

    # If the Python interpreter running this app is UNO-enabled, set None to
    # the following parameter. Else, specify the path to a UNO-enabled
    # interpreter. On Ubuntu, /usr/bin/python3 is UNO-enabled.
    unoEnabledPython = '/usr/bin/python3'
    # On what port does LibreOffice run ?
    libreOfficePort = 2002
    # Monitoring configuration. Update this instance (whose class is in
    # appy.gen.monitoring)) for changing the default configuration.
    monitoring = Monitoring()
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
import os.path, time
from appy.fields.file import FileInfo
from appy.shared import utils as sutils

# ------------------------------------------------------------------------------
class Migrator:
    '''This class is responsible for performing migrations, when, on
       installation, we've detected a new Appy version.'''
    def __init__(self, installer):
        self.installer = installer
        self.logger = installer.logger
        self.app = installer.app
        self.tool = self.app.config.appy()

    @staticmethod
    def migrateBinaryFields(obj):
        '''Ensures all file and frozen pod fields on p_obj are FileInfo
           instances.'''
        migrated = 0 # Count the number of migrated fields
        for field in obj.fields:
            if field.type == 'File':
                oldValue = getattr(obj, field.name)
                if oldValue and not isinstance(oldValue, FileInfo):
                    # A legacy File object. Convert it to a FileInfo instance
                    # and extract the binary to the filesystem.
                    setattr(obj, field.name, oldValue)
                    migrated += 1
            elif field.type == 'Pod':
                frozen = getattr(obj.o.aq_base, field.name, None)
                if frozen:
                    # Dump this file on disk
                    tempFolder = sutils.getOsTempFolder()
                    fmt = os.path.splitext(frozen.filename)[1][1:]
                    fileName = os.path.join(tempFolder,
                                            '%f.%s' % (time.time(), fmt))
                    f = file(fileName, 'wb')
                    if frozen.data.__class__.__name__ == 'Pdata':
                        # The file content is splitted in several chunks
                        f.write(frozen.data.data)
                        nextPart = frozen.data.next
                        while nextPart:
                            f.write(nextPart.data)
                            nextPart = nextPart.next
                    else:
                        # Only one chunk
                        f.write(frozen.data)
                    f.close()
                    f = file(fileName)
                    field.freeze(obj, template=field.template[0], format=fmt,
                                 noSecurity=True, upload=f,
                                 freezeOdtOnError=False)
                    f.close()
                    # Remove the legacy in-zodb file object
                    setattr(obj.o, field.name, None)
                    migrated += 1
        return migrated

    def migrateTo_0_9_0(self):
        '''Migrates this DB to Appy 0.9.x'''
        # Put all binaries to the filesystem
        tool = self.tool
        tool.log('Migrating binary fields...')
        context = {'migrate': self.migrateBinaryFields, 'nb': 0}
        for className in tool.o.getAllClassNames():
            tool.compute(className, context=context, noSecurity=True,
                         expression="ctx['nb'] += ctx['migrate'](obj)")
        tool.log('Migrated %d binary field(s).' % context['nb'])

    def run(self, force=False):
        '''Executes a migration when relevant, or do it for sure if p_force is
           True.'''
        appyVersion = self.tool.appyVersion
        # appyVersion being None simply means that we are creating a new DB
        if force or (appyVersion and (appyVersion < '0.9.0')):
            # Migration is required
            self.logger.info('Appy version (DB) is %s' % appyVersion)
            startTime = time.time()
            self.migrateTo_0_9_0()
            stopTime = time.time()
            elapsed = (stopTime-startTime) / 60.0
            self.logger.info('Migration done in %d minute(s).' % elapsed)
# ------------------------------------------------------------------------------
